from typing import Any, Dict, List, Mapping, Optional, Tuple

import torch

from ...util.string import StringTransform, StringTransformations
from ..hf_hub.conversion import (
    CommonCuratedToHFConverters,
    CommonHFKeys,
    HFConfigKey,
    HFConfigKeyDefault,
    HFSpecificConfig,
    config_from_hf,
    config_to_hf,
)
from ..module import DecoderModule
from .config import MPTConfig

# Order-dependent.
COMMON_HF_PARAM_KEY_TRANSFORMS: List[StringTransform] = [
    StringTransformations.sub("transformer", "decoder"),
    StringTransformations.sub("blocks", "layers"),
    # Attention blocks.
    StringTransformations.sub(".attn", ".mha"),
    StringTransformations.sub(".Wqkv", ".input"),
    StringTransformations.sub(".out_proj", ".output"),
    # Pointwise feedforward.
    StringTransformations.sub(".up_proj", ".intermediate"),
    StringTransformations.sub("ffn.down_proj", "ffn.output"),
    # Layer norms.
    StringTransformations.sub(".norm_1", ".attn_input_layer_norm"),
    StringTransformations.sub(".norm_2", ".ffn_input_layer_norm"),
    StringTransformations.sub("norm_f.", "output_layer_norm."),
    # Embeddings.
    StringTransformations.sub("wte.", "embeddings.piece_embeddings."),
]

DECODER_HF_PARAM_KEY_TRANSFORMS = [
    StringTransformations.remove_prefix("transformer.", reversible=False)
] + COMMON_HF_PARAM_KEY_TRANSFORMS
CAUSAL_LM_HF_PARAM_KEY_TRANSFORMS = COMMON_HF_PARAM_KEY_TRANSFORMS


class HFConfigKeys:
    @staticmethod
    def conv_intermediate_width_multiplier(config: MPTConfig) -> float:
        return (
            config.layer.feedforward.intermediate_width
            / config.layer.feedforward.hidden_width
        )

    @staticmethod
    def conv_model_max_length(config: MPTConfig) -> int:
        return config.model_max_length

    @staticmethod
    def conv_use_bias(config: MPTConfig) -> int:
        # This needs to be flipped due to how it's
        # stored in the HF config.
        return not config.layer.feedforward.use_bias

    D_MODEL = HFConfigKey(
        "d_model",
        "hidden_width",
        lambda c: CommonCuratedToHFConverters.hidden_width(c),
    )
    EXPANSION_RATIO = HFConfigKey(
        "expansion_ratio",
        "intermediate_width_multiplier",
        lambda c: HFConfigKeys.conv_intermediate_width_multiplier(c),
    )
    MAX_SEQ_LEN = HFConfigKey(
        "max_seq_len",
        "model_max_length",
        lambda c: HFConfigKeys.conv_model_max_length(c),
    )
    N_LAYERS = HFConfigKey(
        "n_layers",
        "n_hidden_layers",
        lambda c: CommonCuratedToHFConverters.n_hidden_layers(c),
    )
    N_HEADS = HFConfigKey(
        "n_heads",
        "n_attention_heads",
        lambda c: CommonCuratedToHFConverters.n_attention_heads_uniform(c),
    )
    NO_BIAS = HFConfigKey(
        "no_bias",
        ("use_bias", lambda v: not v),
        lambda c: HFConfigKeys.conv_use_bias(c),
    )
    LAYER_NORM_EPSILON = HFConfigKey(
        "layer_norm_epsilon",
        "layer_norm_eps",
        lambda c: CommonCuratedToHFConverters.layer_norm_eps(c),
    )


HF_CONFIG_KEYS: List[Tuple[HFConfigKey, Optional[HFConfigKeyDefault]]] = [
    (CommonHFKeys.DTYPE, HFConfigKeyDefault("bfloat16")),
    (HFConfigKeys.D_MODEL, None),
    (HFConfigKeys.EXPANSION_RATIO, None),
    (HFConfigKeys.MAX_SEQ_LEN, None),
    (HFConfigKeys.N_LAYERS, None),
    (HFConfigKeys.N_HEADS, None),
    (HFConfigKeys.NO_BIAS, None),
    (CommonHFKeys.VOCAB_SIZE, None),
    (CommonHFKeys.ATTENTION_PROBS_DROPOUT_PROB, HFConfigKeyDefault(0.0)),
    (CommonHFKeys.HIDDEN_DROPOUT_PROB, HFConfigKeyDefault(0.0)),
    (HFConfigKeys.LAYER_NORM_EPSILON, HFConfigKeyDefault(1e-5)),
]

HF_SPECIFIC_CONFIG_DECODER = HFSpecificConfig(
    architectures=["MptModel"], model_type="mpt"
)
HF_SPECIFIC_CONFIG_CAUSAL_LM = HFSpecificConfig(
    architectures=["MptForCausalLM"], model_type="mpt"
)


def _config_from_hf(hf_config: Mapping[str, Any]) -> MPTConfig:
    kwargs = config_from_hf("MPT", hf_config, HF_CONFIG_KEYS)
    return MPTConfig(**kwargs)


def _config_to_hf(cls, curated_config: MPTConfig) -> Dict[str, Any]:
    out = config_to_hf(curated_config, [k for k, _ in HF_CONFIG_KEYS])
    if issubclass(cls, DecoderModule):
        return HF_SPECIFIC_CONFIG_DECODER.merge(out)
    else:
        return HF_SPECIFIC_CONFIG_CAUSAL_LM.merge(out)



import os
import warnings
from tempfile import NamedTemporaryFile
from typing import IO, Any, AnyStr, Dict, Iterator, List, Optional, Union

import huggingface_hub
from huggingface_hub import CommitOperationAdd, HfApi
from huggingface_hub.utils import (
    EntryNotFoundError,
    RepositoryNotFoundError,
    RevisionNotFoundError,
)
from requests import HTTPError, ReadTimeout  # type: ignore
from typing_extensions import Buffer

from ..repository.file import LocalFile, RepositoryFile
from .repository import Repository
from .transaction import TransactionContext


class HfHubFile(RepositoryFile):
    """
    Wraps either a remote file on a Hugging Face Hub repository
    or a local file in the Hugging Face cache.
    """

    def __init__(self, repo: "HfHubRepository", path: str):
        """
        Construct a Hugging Face file representation.

        :param repo:
            The parent repository.
        :param path:
            The path of the remote file in the repository.
        """
        super().__init__()

        self._repo = repo
        self._remote_path = path
        self._cached_path = lookup_file_in_cache(repo.name, repo.revision, path)

    def _validate_model_repo(self):
        api = HfApi()
        # This will raise RepositoryNotFoundError/RevisionNotFoundError upon failure.
        _ = api.model_info(repo_id=self._repo.name, revision=self._repo.revision)

    def _download_to_cache(self) -> LocalFile:
        local_file = LocalFile(
            hf_hub_download(
                repo_id=self._repo.name,
                filename=self._remote_path,
                revision=self._repo.revision,
            )
        )
        self._cached_path = local_file.path
        return local_file

    def open(self, mode: str = "rb", encoding: Optional[str] = None) -> IO:
        try:
            if "r" in mode:
                # Attempt to download the file if we want to read from it.
                local_file = self._download_to_cache()
                return local_file.open(mode=mode, encoding=encoding)
            elif "w" in mode:
                # Return a proxy for the remote file.
                self._validate_model_repo()
                return UploadStagingBuffer(
                    self._repo, self._remote_path, mode=mode, encoding=encoding
                )
            else:
                raise OSError(
                    f"Unsupported mode '{mode}' for opening Hugging Face Hub file"
                )
        except (
            EntryNotFoundError,
            RepositoryNotFoundError,
            RevisionNotFoundError,
        ) as e:
            raise FileNotFoundError(
                f"File not found: {self._repo.pretty_path(self._remote_path)}"
            ) from e
        except Exception as e:
            raise OSError(
                f"File could not be opened: {self._repo.pretty_path(self._remote_path)}"
            ) from e

    @property
    def path(self) -> Optional[str]:
        return self._cached_path

    def exists(self) -> bool:
        try:
            return self._download_to_cache().exists()
        except:
            return False


class HfHubRepository(Repository):
    """
    Hugging Face Hub repository.
    """

    def __init__(self, name: str, *, revision: str = "main"):
        """
        :param name:
            Name of the repository on Hugging Face Hub.
        :param revision:
            Source repository revision. Can either be a branch name
            or a SHA hash of a commit.
        """
        super().__init__()
        self.name = name
        self.revision = revision

    def file(self, path: str) -> RepositoryFile:
        return HfHubFile(repo=self, path=path)

    def pretty_path(self, path: Optional[str] = None) -> str:
        if not path:
            return f"{self.name} (revision: {self.revision})"
        return f"{self.name}/{path} (revision: {self.revision})"

    def transaction(self) -> TransactionContext:
        return HfHubTransactionContext(self)


class HfHubTransactionContext(TransactionContext):
    """
    Hugging Face Hub transaction context manager. Files opened
    for writing are backed by temporary files on the host and
    uploaded to the Hub when the transaction completes successfully.
    """

    def __init__(self, repo: HfHubRepository):
        """
        :param repo:
            The parent repository.
        """
        super().__init__()
        self._repo = repo
        self._file_mappings: Dict[str, IO] = (
            {}
        )  # Maps remote file paths to local temporary files

    def open(self, path: str, mode: str, encoding: Optional[str] = None) -> IO:
        if path in self._file_mappings:
            raise OSError(f"Path '{path}' has already been opened")
        elif "w" not in mode:
            raise OSError(
                f"File open mode '{mode}' is not supported inside a transaction - "
                "files can only be opened for writing"
            )

        temp_file = NamedTemporaryFile(mode, encoding=encoding, delete=False)
        self._file_mappings[path] = temp_file
        return temp_file

    @property
    def repo(self) -> Repository:
        return self._repo

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        try:
            if exc_type is None:
                self._upload_temp_files()
        finally:
            self._release_temp_files()

    def _upload_temp_files(self):
        if len(self._file_mappings) == 0:
            return

        # Flush buffers.
        for temp_file in self._file_mappings.values():
            if not temp_file.closed:
                temp_file.flush()

        # Collate all files as a single commit.
        commit_ops = [
            CommitOperationAdd(path_in_repo=remote_path, path_or_fileobj=temp_file.name)
            for remote_path, temp_file in self._file_mappings.items()
        ]
        filenames = "\n".join(self._file_mappings.keys())
        commit_message = f"Uploading files:\n{filenames}"

        api = HfApi()
        _ = api.create_commit(
            repo_id=self._repo.name,
            revision=self._repo.revision,
            operations=commit_ops,
            commit_message=commit_message,
            run_as_future=False,
        )

    def _release_temp_files(self):
        for temp_file in self._file_mappings.values():
            try:
                if not temp_file.closed:
                    temp_file.close()
            finally:
                os.remove(temp_file.name)


class UploadStagingBuffer(IO):
    """
    Represents a local temporary file buffer that gets uploaded
    to a repository when it's closed.
    """

    def __init__(
        self,
        repo: HfHubRepository,
        remote_path: str,
        *,
        mode: str,
        encoding: Optional[str],
    ):
        super().__init__()
        self._repo = repo
        self._remote_path = remote_path
        self._temp_file = NamedTemporaryFile(mode, encoding=encoding, delete=False)

    def _upload(self):
        try:
            api = HfApi()
            result = api.upload_file(
                path_in_repo=self._remote_path,
                path_or_fileobj=self._temp_file,
                repo_id=self._repo.name,
                revision=self._repo.revision,
                run_as_future=False,
            )
        except Exception as e:
            raise OSError(
                f"Failed to upload to remote file on Hugging Face Hub @ {self._repo.pretty_path(self._remote_path)}.\n"
                f"Error: {e}"
            )

    # IO overrides follow.

    def __iter__(self) -> Iterator[Any]:
        return self._temp_file.__iter__()

    def __next__(self) -> Any:
        return self._temp_file.__next__()

    @property
    def mode(self) -> str:
        return self._temp_file.mode

    @property
    def name(self) -> str:
        return self._temp_file.name

    def close(self) -> None:
        return self._temp_file.close()

    @property
    def closed(self) -> bool:
        return self._temp_file.closed

    def fileno(self) -> int:
        return self._temp_file.fileno()

    def flush(self) -> None:
        return self._temp_file.flush()

    def isatty(self) -> bool:
        return self._temp_file.isatty()

    def read(self, n: int = -1) -> AnyStr:
        return self._temp_file.read(n)

    def readable(self) -> bool:
        return self._temp_file.readable()

    def readline(self, limit: int = -1) -> AnyStr:
        return self._temp_file.readline(limit)

    def readlines(self, hint: int = -1) -> List[AnyStr]:
        return self._temp_file.readlines(hint)

    def seek(self, offset: int, whence: int = 0) -> int:
        return self._temp_file.seek(offset, whence)

    def seekable(self) -> bool:
        return self._temp_file.seekable()

    def tell(self) -> int:
        return self._temp_file.tell()

    def truncate(self, size: int = None) -> int:  # type:ignore
        return self._temp_file.truncate(size)

    def writable(self) -> bool:
        return self._temp_file.writable()

    def write(self, s: Union[Any, Buffer, str]) -> int:
        return self._temp_file.write(s)

    def writelines(self, lines: List[AnyStr]) -> None:  # type:ignore
        self._temp_file.writelines(lines)

    def __enter__(self):
        return self._temp_file

    def __exit__(self, exc_type, exc_value, exc_tb):
        try:
            if exc_type is not None:
                self._upload()
        finally:
            os.remove(self._temp_file.name)


def hf_hub_download(repo_id: str, filename: str, revision: str) -> str:
    """
    Resolve the provided filename and repository to a local file path. If the file
    is not present in the cache, it will be downloaded from the Hugging Face Hub.

    :param repo_id:
        Identifier of the source repository on Hugging Face Hub.
    :param filename:
        Name of the file in the source repository to download.
    :param revision:
        Source repository revision. Can either be a branch name
        or a SHA hash of a commit.
    :returns:
        Resolved absolute filepath.
    """

    # The HF Hub library's `hf_hub_download` function will always attempt to connect to the
    # remote repo and fetch its metadata even if it's locally cached (in order to update the
    # out-of-date artifacts in the cache). This can occasionally lead to `HTTPError/ReadTimeout`s if the
    # remote host is unreachable. Instead of failing loudly, we'll add a fallback that checks
    # the local cache for the artifacts and uses them if found.
    try:
        resolved = huggingface_hub.hf_hub_download(
            repo_id=repo_id, filename=filename, revision=revision
        )
    except (HTTPError, ReadTimeout) as e:
        # Attempt to check the cache.
        resolved = lookup_file_in_cache(repo_id, revision, filename)
        if resolved is None:
            # Not found, rethrow.
            raise e
        else:
            warnings.warn(
                f"Couldn't reach Hugging Face Hub; using cached artifact for '{repo_id}@{revision}:{filename}'"
            )
    return resolved


def lookup_file_in_cache(repo_id: str, revision: str, filename: str) -> Optional[str]:
    """
    Look up the file in the local Hugging Face Hub cache.

    :param repo_id:
        Identifier of the source repository on Hugging Face Hub.
    :param revision:
        Source repository revision. Can either be a branch name
        or a SHA hash of a commit.
    :param filename:
        Name of the file in the source repository to lookup in the cache.
    :returns:
        Local file path if found in cache, ``None`` otherwise.
    """

    resolved = huggingface_hub.try_to_load_from_cache(
        repo_id=repo_id, filename=filename, revision=revision
    )
    if resolved is None or resolved is huggingface_hub._CACHED_NO_EXIST:
        return None
    else:
        return resolved



import pytest

from curated_transformers.tokenizers.legacy import LlamaTokenizer

from ...compat import has_hf_transformers
from ..util import compare_tokenizer_outputs_with_hf_tokenizer


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_from_hf_hub_equals_hf_tokenizer(sample_texts):
    # OpenLlama does not provide a fast tokenizer. If we ask for a fast
    # tokenizer, the slow tokenizer gets converted, which takes too much
    # time for CI. So, we use the slow tokenizer instead.
    compare_tokenizer_outputs_with_hf_tokenizer(
        sample_texts,
        "openlm-research/open_llama_3b",
        LlamaTokenizer,
        with_hf_fast=False,
        pad_token="<unk>",
    )



import pytest

from curated_transformers.models.camembert.encoder import CamemBERTEncoder

from ...compat import has_hf_transformers, has_torch_compile
from ...conftest import TORCH_DEVICES
from ..util import (
    JITMethod,
    assert_encoder_output_equals_hf,
    assert_model_hf_serialization_roundtrip,
)


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
@pytest.mark.parametrize("with_torch_sdp", [False, True])
def test_encoder(torch_device, with_torch_sdp):
    assert_encoder_output_equals_hf(
        CamemBERTEncoder,
        "explosion-testing/camembert-test",
        torch_device,
        with_torch_sdp=with_torch_sdp,
    )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.skipif(not has_torch_compile, reason="requires torch.compile")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
@pytest.mark.parametrize("with_torch_sdp", [False, True])
def test_encoder_with_torch_compile(torch_device, with_torch_sdp):
    assert_encoder_output_equals_hf(
        CamemBERTEncoder,
        "explosion-testing/camembert-test",
        torch_device,
        jit_method=JITMethod.TorchCompile,
        with_torch_sdp=with_torch_sdp,
    )


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
def test_encoder_hf_serializtion_roundtrip(torch_device):
    assert_model_hf_serialization_roundtrip(
        CamemBERTEncoder, "explosion-testing/camembert-test", torch_device
    )



from typing import Dict, Optional, Type, cast

import torch

from ..models.auto_model import AutoModel
from ..quantization.bnb.config import BitsAndBytesConfig
from .dolly_v2 import DollyV2Generator
from .falcon import FalconGenerator
from .generator_wrapper import GeneratorWrapper
from .hf_hub import FromHF
from .llama import LlamaGenerator
from .mpt import MPTGenerator

# For the time being, we enable support for a generator on a case-by-case basis.
# In the future we might defer all unknown generators to DefaultGenerator.
GENERATOR_MAP: Dict[str, Type[FromHF]] = {
    "dolly-v2": DollyV2Generator,
    "falcon": FalconGenerator,
    "llama": LlamaGenerator,
    "mpt": MPTGenerator,
}


class AutoGenerator(AutoModel[GeneratorWrapper]):
    """Causal LM generator loaded from the Hugging Face Model Hub.

    .. attention::
            This class can currently only be used with the following models:

            - Models based on Dolly v2 (contain ``dolly-v2`` in the name).
            - Models based on Falcon (contain ``falcon`` in the name).
            - Models based on Llama (contain ``llama`` in the name).
            - Models based on MPT (contain ``mpt`` in the name).
    """

    @classmethod
    def from_hf_hub_to_cache(
        cls,
        *,
        name: str,
        revision: str = "main",
    ):
        generator_cls = _resolve_generator_class(name)
        generator_cls.from_hf_hub_to_cache(name=name, revision=revision)

    @classmethod
    def from_hf_hub(
        cls,
        *,
        name: str,
        revision: str = "main",
        device: Optional[torch.device] = None,
        quantization_config: Optional[BitsAndBytesConfig] = None,
    ) -> GeneratorWrapper:
        # We need to match the name of the model directly as our
        # generators bundle model-specific prompts that are specific
        # to certain fine-tuned models.
        generator_cls = _resolve_generator_class(name)
        generator = cast(
            GeneratorWrapper,
            generator_cls.from_hf_hub(
                name=name,
                revision=revision,
                device=device,
                quantization_config=quantization_config,
            ),
        )

        assert isinstance(generator, GeneratorWrapper)
        return generator


def _resolve_generator_class(name: str) -> Type[FromHF]:
    for substring, generator_cls in GENERATOR_MAP.items():
        if substring in name.lower():
            return generator_cls

    supported_models = list(sorted(GENERATOR_MAP.keys()))
    raise ValueError(
        f"Unsupported generator model `{name}`. "
        f"Supported model variants: `{supported_models}`"
    )



from typing import Any, Dict, Mapping, Optional, Tuple, Type, TypeVar

import torch
from torch import Tensor
from torch.nn import Dropout, LayerNorm

from ...layers.attention import AttentionMask
from ...layers.transformer import (
    EmbeddingDropouts,
    EmbeddingLayerNorms,
    TransformerEmbeddings,
)
from ..hf_hub import FromHF
from ..hf_hub.conversion import state_dict_from_hf, state_dict_to_hf
from ..module import EncoderModule
from ..output import ModelOutput
from ._hf import HF_PARAM_KEY_TRANSFORMS, _config_from_hf, _config_to_hf
from .config import ALBERTConfig
from .layer_group import ALBERTLayerGroup

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="ALBERTEncoder")


class ALBERTEncoder(EncoderModule[ALBERTConfig], FromHF[ALBERTConfig]):
    """
    ALBERT (`Lan et al., 2022`_) encoder.

    .. _Lan et al., 2022: https://arxiv.org/abs/1909.11942
    """

    def __init__(self, config: ALBERTConfig, *, device: Optional[torch.device] = None):
        """
        Construct an ALBERT encoder.

        :param config:
            Encoder configuration.
        :param device:
            Device to which the module is to be moved.
        :returns:
            The encoder.
        """
        super().__init__(config)

        self.max_seq_len = config.model_max_length
        self.n_hidden_layers = config.layer.n_hidden_layers
        n_hidden_groups = config.layer.n_hidden_groups

        if self.n_hidden_layers % n_hidden_groups != 0:
            raise ValueError(
                f"The number of hidden layers ({self.n_hidden_layers}) in the "
                "ALBERT encoder must be divisable by number of hidden groups "
                f"({n_hidden_groups})"
            )

        self.embeddings = TransformerEmbeddings(
            dropouts=EmbeddingDropouts(
                embed_output_dropout=Dropout(config.embedding.dropout_prob)
            ),
            embedding_width=config.embedding.embedding_width,
            hidden_width=config.layer.feedforward.hidden_width,
            layer_norms=EmbeddingLayerNorms(
                embed_output_layer_norm=LayerNorm(
                    config.embedding.embedding_width, config.embedding.layer_norm_eps
                )
            ),
            n_pieces=config.embedding.n_pieces,
            n_positions=config.embedding.n_positions,
            n_types=config.embedding.n_types,
            device=device,
        )

        # Parameters are shared by groups of layers.
        self.groups = torch.nn.ModuleList(
            [
                ALBERTLayerGroup(config.layer, device=device)
                for _ in range(n_hidden_groups)
            ]
        )

    def forward(
        self,
        piece_ids: Tensor,
        attention_mask: AttentionMask,
        *,
        type_ids: Optional[Tensor] = None,
        positions: Optional[Tensor] = None,
    ) -> ModelOutput:
        embeddings = self.embeddings(piece_ids, positions=positions, type_ids=type_ids)
        layer_output = embeddings

        layers_per_group = self.n_hidden_layers // len(self.groups)

        layer_outputs = []
        for group in self.groups:
            for _ in range(layers_per_group):
                layer_output, _ = group(layer_output, attention_mask=attention_mask)
                layer_outputs.append(layer_output)

        return ModelOutput(all_outputs=[embeddings, *layer_outputs])

    @classmethod
    def is_supported(cls: Type[Self], config: Dict[str, Any]) -> bool:
        return config.get("model_type") == "albert"

    @classmethod
    def state_dict_from_hf(
        cls: Type[Self], params: Mapping[str, Tensor]
    ) -> Mapping[str, Tensor]:
        return state_dict_from_hf(params, HF_PARAM_KEY_TRANSFORMS)

    @classmethod
    def state_dict_to_hf(
        cls: Type[Self], params: Mapping[str, Tensor]
    ) -> Mapping[str, Tensor]:
        return state_dict_to_hf(params, HF_PARAM_KEY_TRANSFORMS)

    @classmethod
    def config_from_hf(cls, hf_config: Mapping[str, Any]) -> ALBERTConfig:
        return _config_from_hf(hf_config)

    @classmethod
    def config_to_hf(cls, curated_config: ALBERTConfig) -> Mapping[str, Any]:
        return _config_to_hf(curated_config)

    @classmethod
    def from_hf_config(
        cls: Type[Self],
        *,
        hf_config: Any,
        device: Optional[torch.device] = None,
    ) -> Self:
        config = cls.config_from_hf(hf_config)
        return cls(config, device=device)



import pytest

from curated_transformers.models.hf_hub.conversion import CommonHFKeys
from curated_transformers.models.mpt._hf import HFConfigKeys
from curated_transformers.models.mpt.decoder import MPTDecoder

from ...compat import has_hf_transformers, has_torch_compile
from ...conftest import TORCH_DEVICES
from ..util import (
    JITMethod,
    assert_decoder_output_equals_hf,
    assert_model_hf_serialization_roundtrip,
)


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
@pytest.mark.parametrize("with_torch_sdp", [False, True])
def test_decoder(torch_device, with_torch_sdp):
    assert_decoder_output_equals_hf(
        MPTDecoder,
        "explosion-testing/mpt-test",
        torch_device,
        # HF model does not support position IDs.
        with_positions=False,
        with_torch_sdp=with_torch_sdp,
    )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.skipif(not has_torch_compile, reason="requires torch.compile")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
@pytest.mark.parametrize("with_torch_sdp", [False, True])
def test_decoder_with_torch_compile(torch_device, with_torch_sdp):
    # NOTE: testing with cache/positions fails. All outputs seems ok until
    # the concatenation of the vectors with/without rotary embeddings:
    #
    # https://github.com/explosion/curated-transformers/blob/6de9b9828b1d6f8f52dba8b87b4e00a9732a2d5f/curated_transformers/layers/embeddings.py#L308
    assert_decoder_output_equals_hf(
        MPTDecoder,
        "explosion-testing/mpt-test",
        torch_device,
        with_cache=False,
        with_positions=False,
        jit_method=JITMethod.TorchCompile,
        with_torch_sdp=with_torch_sdp,
    )


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
def test_decoder_hf_serializtion_roundtrip(torch_device):
    assert_model_hf_serialization_roundtrip(
        MPTDecoder,
        "explosion-testing/mpt-test",
        torch_device,
        optional_hf_config_keys={
            HFConfigKeys.LAYER_NORM_EPSILON.name,
            CommonHFKeys.ATTENTION_PROBS_DROPOUT_PROB.name,
            CommonHFKeys.HIDDEN_DROPOUT_PROB.name,
        },
    )



import math
from enum import Enum
from typing import Type

import torch
from torch import Tensor
from torch.nn import Module


class Activation(Enum):
    """
    Activation functions.

    .. _Black et al., 2022: https://arxiv.org/abs/2204.06745
    .. _Hendrycks et al., 2016: https://arxiv.org/abs/1606.08415
    .. _Fukushima, 1969: https://ieeexplore.ieee.org/document/4082265
    """

    #: Rectified Linear Unit (`Fukushima, 1969`_).
    ReLU = "relu"

    #: Gaussian Error Linear Unit (`Hendrycks et al., 2016`_).
    GELU = "gelu"

    #: Gaussian Error Linear Unit (`Hendrycks et al., 2016`_) approximation
    #: used by GPT-NeoX (`Black et al., 2022`_).
    GELUFast = "gelu_fast"

    #: Gaussian Error Linear Unit (`Hendrycks et al., 2016`_) approximation.
    GELUNew = "gelu_new"

    #: Sigmoid Linear Unit (`Hendrycks et al., 2016`_).
    SiLU = "silu"

    @classmethod
    def _missing_(cls, value):
        supported_activations = ", ".join(sorted(v.value for v in cls))
        raise ValueError(
            f"Invalid activation function `{value}`. "
            f"Supported functions: {supported_activations}"
        )

    @property
    def module(self) -> Type[torch.nn.Module]:
        """
        Get the PyTorch module for the activation function.
        """

        # TODO: convert to match when Python 3.10 is the minimum version.
        if self == Activation.ReLU:
            return torch.nn.ReLU
        elif self == Activation.GELU:
            return torch.nn.GELU
        elif self == Activation.GELUFast:
            return GELUFast
        elif self == Activation.GELUNew:
            return GELUNew
        elif self == Activation.SiLU:
            return torch.nn.SiLU
        else:
            raise NotImplementedError


class GELUNew(Module):
    """
    GELU (`Hendrycks et al., 2016`_) approximation, called ``gelu_new`` in many transformer models.

    .. _Hendrycks et al., 2016: https://arxiv.org/abs/1606.08415
    """

    def forward(self, input: Tensor) -> Tensor:
        """
        Apply the GELU activation on the input.

        :param input:
            Input tensor.

            *Shape:* ``(batch_size, seq_len, width)``
        """
        return (
            0.5
            * input
            * (
                1.0
                + torch.tanh(
                    math.sqrt(2.0 / math.pi)
                    * (input + 0.044715 * torch.pow(input, 3.0))
                )
            )
        )


class GELUFast(Module):
    """
    GELU (`Hendrycks et al., 2016`_) approximation used by GPT-NeoX (`Black et al., 2022`_).

    .. _Hendrycks et al., 2016: https://arxiv.org/abs/1606.08415
    .. _Black et al., 2022: https://arxiv.org/abs/2204.06745
    """

    def forward(self, input: Tensor) -> Tensor:
        """
        Apply the GELU activation on the input.

        :param input:
            Input tensor.

            *Shape:* ``(batch_size, seq_len, width)``
        """
        alpha = math.sqrt(2.0 / math.pi)
        beta = 0.044715

        return 0.5 * input * (1.0 + torch.tanh(alpha * (input + beta * input.pow(3))))



import pytest
from huggingface_hub import HfFileSystem

from curated_transformers.models import (
    ALBERTEncoder,
    BERTEncoder,
    CamemBERTEncoder,
    FalconCausalLM,
    FalconDecoder,
    GPTNeoXCausalLM,
    GPTNeoXDecoder,
    RoBERTaEncoder,
    XLMREncoder,
)
from curated_transformers.models.auto_model import (
    AutoCausalLM,
    AutoDecoder,
    AutoEncoder,
)
from curated_transformers.models.llama.causal_lm import LlamaCausalLM
from curated_transformers.models.llama.decoder import LlamaDecoder
from curated_transformers.models.mpt.causal_lm import MPTCausalLM
from curated_transformers.models.mpt.decoder import MPTDecoder
from curated_transformers.repository.fsspec import FsspecArgs


@pytest.fixture
def model_encoder_map():
    return {
        "explosion-testing/bert-test": BERTEncoder,
        "explosion-testing/albert-test": ALBERTEncoder,
        "explosion-testing/roberta-test": RoBERTaEncoder,
        "explosion-testing/camembert-test": CamemBERTEncoder,
        "explosion-testing/xlm-roberta-test": XLMREncoder,
    }


def test_auto_encoder(model_encoder_map):
    for name, encoder_cls in model_encoder_map.items():
        encoder = AutoEncoder.from_hf_hub(name=name)
        assert isinstance(encoder, encoder_cls)

    with pytest.raises(ValueError, match="Unsupported model type"):
        AutoEncoder.from_hf_hub(name="explosion-testing/falcon-test")


@pytest.mark.slow
def test_auto_encoder_fsspec(model_encoder_map):
    for name, encoder_cls in model_encoder_map.items():
        # The default revision is 'main', but we pass it anyway to test
        # that the function acceps fsspec_args.
        encoder = AutoEncoder.from_fsspec(
            fs=HfFileSystem(), model_path=name, fsspec_args=FsspecArgs(revision="main")
        )
        assert isinstance(encoder, encoder_cls)


def test_auto_decoder():
    model_decoder_map = {
        "explosion-testing/falcon-test": FalconDecoder,
        "explosion-testing/llama2-kv-sharing": LlamaDecoder,
        "explosion-testing/mpt-test": MPTDecoder,
        "trl-internal-testing/tiny-random-GPTNeoXForCausalLM": GPTNeoXDecoder,
    }

    for name, decoder_cls in model_decoder_map.items():
        decoder = AutoDecoder.from_hf_hub(name=name)
        assert isinstance(decoder, decoder_cls)

    with pytest.raises(ValueError, match="Unsupported model type"):
        AutoDecoder.from_hf_hub(name="trl-internal-testing/tiny-random-GPT2Model")


def test_auto_causal_lm():
    model_causallm_map = {
        "explosion-testing/falcon-test": FalconCausalLM,
        "explosion-testing/llama2-kv-sharing": LlamaCausalLM,
        "explosion-testing/mpt-test": MPTCausalLM,
        "trl-internal-testing/tiny-random-GPTNeoXForCausalLM": GPTNeoXCausalLM,
    }

    for name, causal_lm_cls in model_causallm_map.items():
        causal_lm = AutoCausalLM.from_hf_hub(name=name)
        assert isinstance(causal_lm, causal_lm_cls)

    with pytest.raises(ValueError, match="Unsupported model type"):
        AutoCausalLM.from_hf_hub(name="trl-internal-testing/tiny-random-GPT2Model")



import dataclasses
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Mapping, Optional, Tuple, Union

import torch
from torch import Tensor

from ...layers import Activation
from ...util.string import StringTransform
from ..config import TransformerConfig


@dataclass
class HFSpecificConfig:
    """
    Configuration options that must always be present in
    a Hugging Face compatible config file.

    The transformer version is the same for all of our
    models.

    :param architectures:
        Model-specific Hugging Face classes that can load the model.
    :param model_type:
        Hugging Face model type.
    :transformers_version:
        Minimum version of the ``transformers`` required to load the model.
    """

    architectures: List[str]
    model_type: str
    transformers_version: str = "4.33.0"

    def merge(self, hf_config: Mapping[str, Any]) -> Dict[str, Any]:
        """
        Merges the keys with a Hugging Face configuration.

        :param hf_config:
            Hugging Face model configuration.
        :returns:
            Merged configuration.
        """
        out = dict(hf_config)
        out.update(dataclasses.asdict(self))
        return out


@dataclass
class HFConfigKey:
    """
    Metadata for a Hugging Face configuration key that describes
    how it must be (de-)serialized.

    :param name:
        Name of the key in the Hugging Face config file.
    :param mapping_to_curated_config:
        Either a string or a tuple of a string and a callable.
        The string is the name of the keyword argument passed
        to the constructor of the Curated Transformers config and
        the callable accepts the value of the key and returns a
        converted value.
    :param mapping_from_curated_config:
        A callable that accepts the config instance and returns
        a converted value.
    """

    name: str
    mapping_to_curated_config: Union[str, Tuple[str, Callable[[Any], Any]]]
    mapping_from_curated_config: Callable[[Any], Any]

    @property
    def curated_config_kwarg(self) -> str:
        """
        Returns the name of the Curated Transformer config keyword
        argument.
        """
        return (
            self.mapping_to_curated_config
            if isinstance(self.mapping_to_curated_config, str)
            else self.mapping_to_curated_config[0]
        )

    def get_kwarg(self, kwargs: Dict[str, Any]) -> Any:
        """
        Looks up the value of the key in the given keyword
        argument dictionary.

        :param kwargs:
            Keyword argument dictionary.
        :returns:
            Value of the key.
        """
        return kwargs[self.curated_config_kwarg]

    def set_kwarg(self, value: Union["HFConfigKey", Any], kwargs: Dict[str, Any]):
        """
        Converts and sets the value of the keyword argument in
        the given dictionary.

        :param value:
            Either a config key instance or a value. In the case
            of the former, its value is looked up in the keyword
            argument dictionary and saved as the calling config
            key's value as-is. In the case of the latter, the
            value is first converted (if possible) before being
            set.
        :param kwargs:
            Keyword argument dictionary.
        """
        if isinstance(value, HFConfigKey):
            kwargs[self.curated_config_kwarg] = value.get_kwarg(kwargs)
            return

        if isinstance(self.mapping_to_curated_config, tuple):
            curated, ctor = self.mapping_to_curated_config
        else:
            curated = self.mapping_to_curated_config
            ctor = lambda x: x

        kwargs[curated] = ctor(value)

    def remove_kwarg(self, kwargs: Dict[str, Any]):
        """
        Removes the key the given keyword argument dictionary.

        :param kwargs:
            Keyword argument dictionary.
        """
        kwargs.pop(self.curated_config_kwarg)


@dataclass
class HFConfigKeyDefault:
    """
    Represents a default value for a Hugging Face configuration
    key. Required to pass ``None`` as a valid default value.

    :param value:
        The default value for the key.
    """

    value: Any


class CommonCuratedToHFConverters:
    """
    Common functions to convert Curated Transformer config
    values to a compatible Hugging Face config format.
    """

    @staticmethod
    def n_attention_heads_uniform(config: TransformerConfig) -> int:
        assert (
            config.layer.attention.n_key_value_heads
            == config.layer.attention.n_query_heads
        )
        return config.layer.attention.n_key_value_heads

    @staticmethod
    def attention_dropout(config: TransformerConfig) -> float:
        return config.layer.attention.dropout_prob

    @staticmethod
    def activation(config: TransformerConfig) -> str:
        return config.layer.feedforward.activation.value

    @staticmethod
    def dtype(config: TransformerConfig) -> str:
        return str(torch.float32).split(".")[1]

    @staticmethod
    def embedding_width(config: TransformerConfig) -> int:
        return config.embedding.embedding_width

    @staticmethod
    def hidden_width(config: TransformerConfig) -> int:
        return config.layer.feedforward.hidden_width

    @staticmethod
    def intermediate_width(config: TransformerConfig) -> int:
        return config.layer.feedforward.intermediate_width

    @staticmethod
    def layer_norm_eps(config: TransformerConfig) -> float:
        return config.layer.layer_norm_eps

    @staticmethod
    def hidden_dropout(config: TransformerConfig) -> float:
        return config.layer.dropout_prob

    @staticmethod
    def n_pieces(config: TransformerConfig) -> int:
        return config.embedding.n_pieces

    @staticmethod
    def n_types(config: TransformerConfig) -> Optional[int]:
        return config.embedding.n_types

    @staticmethod
    def n_hidden_layers(config: TransformerConfig) -> int:
        return config.layer.n_hidden_layers

    @staticmethod
    def n_positions(config: TransformerConfig) -> Optional[int]:
        return config.embedding.n_positions


class CommonHFToCuratedConverters:
    """
    Common functions to convert Hugging Face config
    values to a compatible Curated config format.
    """

    @staticmethod
    def dtype(serialized_dtype_str: str) -> Optional[torch.dtype]:
        serialized_dtype = getattr(torch, serialized_dtype_str, None)
        if not isinstance(serialized_dtype, torch.dtype):
            raise ValueError(f"Invalid torch dtype `{serialized_dtype_str}`")
        return serialized_dtype


class CommonHFKeys:
    """
    Common Hugging Face config keys.
    """

    ATTENTION_PROBS_DROPOUT_PROB = HFConfigKey(
        "attention_probs_dropout_prob",
        "attention_probs_dropout_prob",
        # Workaround for Python 3.8 limitation that doesn't allow
        # passing/calling static methods as without a class bound.
        lambda c: CommonCuratedToHFConverters.attention_dropout(c),
    )
    DTYPE = HFConfigKey(
        "torch_dtype",
        ("dtype", lambda h: CommonHFToCuratedConverters.dtype(h)),
        lambda c: CommonCuratedToHFConverters.dtype(c),
    )
    EMBEDDING_SIZE = HFConfigKey(
        "embedding_size",
        "embedding_width",
        lambda c: CommonCuratedToHFConverters.embedding_width(c),
    )
    HIDDEN_ACT = HFConfigKey(
        "hidden_act",
        ("activation", Activation),
        lambda c: CommonCuratedToHFConverters.activation(c),
    )
    HIDDEN_SIZE = HFConfigKey(
        "hidden_size",
        "hidden_width",
        lambda c: CommonCuratedToHFConverters.hidden_width(c),
    )
    INTERMEDIATE_SIZE = HFConfigKey(
        "intermediate_size",
        "intermediate_width",
        lambda c: CommonCuratedToHFConverters.intermediate_width(c),
    )
    LAYER_NORM_EPS = HFConfigKey(
        "layer_norm_eps",
        "layer_norm_eps",
        lambda c: CommonCuratedToHFConverters.layer_norm_eps(c),
    )
    MAX_POSITION_EMBEDDINGS = HFConfigKey(
        "max_position_embeddings",
        "n_positions",
        lambda c: CommonCuratedToHFConverters.n_positions(c),
    )
    TYPE_VOCAB_SIZE = HFConfigKey(
        "type_vocab_size",
        "n_types",
        lambda c: CommonCuratedToHFConverters.n_types(c),
    )
    VOCAB_SIZE = HFConfigKey(
        "vocab_size",
        "n_pieces",
        lambda c: CommonCuratedToHFConverters.n_pieces(c),
    )
    NUM_ATTENTION_HEADS_UNIFORM = HFConfigKey(
        "num_attention_heads",
        "n_attention_heads",
        lambda c: CommonCuratedToHFConverters.n_attention_heads_uniform(c),
    )
    NUM_HIDDEN_LAYERS = HFConfigKey(
        "num_hidden_layers",
        "n_hidden_layers",
        lambda c: CommonCuratedToHFConverters.n_hidden_layers(c),
    )
    HIDDEN_DROPOUT_PROB = HFConfigKey(
        "hidden_dropout_prob",
        "hidden_dropout_prob",
        lambda c: CommonCuratedToHFConverters.hidden_dropout(c),
    )


def config_from_hf(
    model_name: str,
    hf_config: Mapping[str, Any],
    hf_keys: List[Tuple[HFConfigKey, Optional[HFConfigKeyDefault]]],
) -> Dict[str, Any]:
    """
    Convert Hugging Face configuration keys to keyword arguments for
    Curated Transformers configuration classes.

    :param model_name:
        Model name. Only used in exception messages.
    :param hf_config:
        Hugging Face model configuration.
    :param hf_keys:
        List of Hugging Face configuration key descriptor and their
        defaults. Keys without any defaults are treated as mandatory
        and an error is raised if they are not found in the config.
    :returns:
        Dictionary with keyword arguments.
    """

    missing_keys = tuple(
        sorted(
            set(k.name for k, default in hf_keys if default is None).difference(
                set(hf_config.keys())
            )
        )
    )
    if len(missing_keys) != 0:
        raise ValueError(
            f"Missing keys in Hugging Face {model_name} model config: {missing_keys}"
        )

    kwargs: Dict[str, Any] = {}
    for hf_key, default in hf_keys:
        hf_key.set_kwarg(
            hf_config.get(hf_key.name, default.value if default is not None else None),
            kwargs,
        )
    return kwargs


def config_to_hf(
    curated_config: TransformerConfig,
    hf_keys: List[HFConfigKey],
) -> Dict[str, Any]:
    """
    Convert a Curated Transformers configuration to a compatible
    Hugging Face configuration dictionary.

    :param curated_config:
        Curated Transformers model configuration.
    :param hf_keys:
        List of Hugging Face configuration key descriptors.
    :returns:
        Dictionary of a Hugging Face configuration
    """
    hf_config = {}
    for hf_key in hf_keys:
        value = hf_key.mapping_from_curated_config(curated_config)
        if value is not None:
            hf_config[hf_key.name] = value
    return hf_config


def state_dict_from_hf(
    params: Mapping[str, Tensor], transforms: List[StringTransform]
) -> Mapping[str, Tensor]:
    """
    Apply transformations to a Hugging Face state dict to make it
    compatible with Curated Transformer modules.

    :param params:
        Hugging Face state dict.
    :param transforms:
        List of string transformations for the state dict's keys.
    :returns:
        Transformed state dict.
    """
    out = {}
    for key, param in params.items():
        for transform in transforms:
            key = transform.apply(key)
        out[key] = param
    return out


def state_dict_to_hf(
    params: Mapping[str, Tensor], transforms: List[StringTransform]
) -> Mapping[str, Tensor]:
    """
    Apply transformations to a Curated Transformer state dict to make it
    compatible with Hugging Face modules.

    :param params:
        Curated Transformer state dict.
    :param transforms:
        List of string transformations for the state dict's keys.
        This must be the same transformations that were used to
        convert the original Hugging Face state dict.
    :returns:
        Transformed state dict.
    """
    out = {}
    for key, param in params.items():
        for transform in transforms[::-1]:
            key = transform.revert(key)
        out[key] = param
    return out



from enum import Enum
from typing import Callable, Dict, List, Set, Tuple, Type, Union

import torch
from huggingface_hub import HfFileSystem
from torch import Tensor
from torch.nn import Module

from curated_transformers.layers.attention import AttentionMask, enable_torch_sdp
from curated_transformers.layers.cache import KeyValueCache
from curated_transformers.models.hf_hub import FromHF
from curated_transformers.models.module import (
    CausalLMModule,
    DecoderModule,
    EncoderModule,
    TransformerModule,
)
from curated_transformers.models.output import ModelOutput, ModelOutputWithCache
from curated_transformers.repository.hf_hub import HfHubRepository
from curated_transformers.repository.repository import ModelRepository

from ..compat import transformers
from ..utils import torch_assertclose


class DecoderWithCache(Module):
    def __init__(self, decoder: DecoderModule):
        super().__init__()
        self.inner = decoder

    def forward(
        self,
        piece_ids: Tensor,
        attention_mask: AttentionMask,
        cache: List[KeyValueCache],
    ):
        return self.inner.forward(
            piece_ids=piece_ids,
            attention_mask=attention_mask,
            cache=cache,
            store_cache=True,
        )


class DecoderWithPositions(Module):
    def __init__(self, decoder: DecoderModule):
        super().__init__()
        self.inner = decoder

    def forward(
        self, piece_ids: Tensor, attention_mask: AttentionMask, positions: Tensor
    ):
        return self.inner.forward(
            piece_ids=piece_ids, attention_mask=attention_mask, positions=positions
        )


class JITMethod(Enum):
    Disable = 0
    TorchCompile = 1

    def convert(self, model: Module, with_torch_sdp: bool, *args) -> Tuple[
        Module,
        Callable[[Union[ModelOutput, Dict[str, torch.Tensor]]], Tensor],
    ]:
        with enable_torch_sdp(with_torch_sdp):
            if self == JITMethod.TorchCompile:
                return torch.compile(model), lambda s: s
            else:  # JITMethod.Disable
                return model, lambda s: s


def assert_causal_lm_output_equals_hf(
    model_class: Type[FromHF],
    model_name: str,
    torch_device: torch.device,
    *,
    atol=1e-5,
    rtol=1e-5,
    model_revision: str = "main",
    jit_method: JITMethod = JITMethod.Disable,
    with_torch_sdp=False,
):
    orig_model = model_class.from_hf_hub(
        name=model_name, revision=model_revision, device=torch_device
    )
    orig_model.eval()

    check_params_buffers(orig_model, torch_device)

    hf_model = transformers.AutoModelForCausalLM.from_pretrained(
        model_name,
        revision=model_revision,
    )
    hf_model.eval()
    hf_model.to(torch_device)

    torch.manual_seed(0)
    X_jit = torch.randint(0, hf_model.config.vocab_size, (3, 5), device=torch_device)
    mask_jit = torch.ones_like(X_jit, dtype=torch.bool)
    model, get_output = jit_method.convert(
        orig_model, with_torch_sdp, X_jit, AttentionMask(mask_jit)
    )

    X = torch.randint(0, hf_model.config.vocab_size, (2, 10), device=torch_device)
    mask = torch.ones_like(X, dtype=torch.bool)
    with torch.no_grad():
        Y = get_output(model(X, AttentionMask(mask))).logits
        Y_hf = hf_model(X).logits
    torch_assertclose(Y, Y_hf, atol=atol, rtol=rtol)

    mask_jit = torch.rand_like(X_jit, dtype=torch.float32) < 0.5
    model, get_output = jit_method.convert(
        orig_model, with_torch_sdp, X_jit, AttentionMask(mask_jit)
    )

    mask = torch.rand((2, 10), dtype=torch.float, device=torch_device) < 0.5
    with torch.no_grad():
        Y = get_output(model(X, AttentionMask(mask))).logits * mask.unsqueeze(-1)
        Y_hf = hf_model(X, attention_mask=mask).logits * mask.unsqueeze(-1)
    torch_assertclose(Y, Y_hf, atol=atol, rtol=rtol)


def assert_decoder_output_equals_hf(
    model_class: Type[FromHF],
    model_name: str,
    torch_device: torch.device,
    *,
    atol=1e-5,
    rtol=1e-5,
    model_revision: str = "main",
    trust_remote_code: bool = False,
    with_cache: bool = True,
    with_positions: bool = True,
    with_mask: bool = True,
    jit_method: JITMethod = JITMethod.Disable,
    with_torch_sdp=False,
    check_config=True,
):
    orig_model = model_class.from_hf_hub(
        name=model_name, revision=model_revision, device=torch_device
    )
    orig_model.eval()

    check_params_buffers(orig_model, torch_device)

    hf_model = transformers.AutoModel.from_pretrained(
        model_name, revision=model_revision, trust_remote_code=trust_remote_code
    )
    hf_model.to(torch_device)
    hf_model.eval()

    torch.manual_seed(0)
    X_jit = torch.randint(0, hf_model.config.vocab_size, (3, 5), device=torch_device)
    mask_jit = torch.ones_like(X_jit, dtype=torch.bool)

    model, output = jit_method.convert(
        orig_model, with_torch_sdp, X_jit, AttentionMask(mask_jit)
    )

    X = torch.randint(0, hf_model.config.vocab_size, (2, 10), device=torch_device)
    mask = torch.ones_like(X, dtype=torch.bool)
    with torch.no_grad():
        Y = output(model(X, AttentionMask(mask))).last_hidden_layer_state
        Y_hf = hf_model(X).last_hidden_state

    torch_assertclose(Y, Y_hf, atol=atol, rtol=rtol)

    if with_cache:
        assert_decoder_with_cache_output_equals_hf(
            orig_model, hf_model, torch_device, atol, rtol, jit_method
        )

    if with_positions:
        assert_decoder_with_positions_equals_hf(
            orig_model, hf_model, torch_device, atol, rtol, jit_method
        )

    if with_mask:
        assert_with_mask_output_equals_hf(
            orig_model, hf_model, torch_device, atol, rtol, jit_method
        )

    if check_config and jit_method == JITMethod.Disable:
        assert_model_config(model, Y)


def assert_encoder_output_equals_hf(
    model_class: Type[FromHF],
    model_name: str,
    torch_device: torch.device,
    *,
    atol: float = 1e-5,
    rtol: float = 1e-5,
    jit_method: JITMethod = JITMethod.Disable,
    with_fsspec: bool = False,
    with_torch_sdp: bool = False,
    check_config=True,
):
    if with_fsspec:
        orig_model = model_class.from_fsspec(
            fs=HfFileSystem(), model_path=model_name, device=torch_device
        )
    else:
        orig_model = model_class.from_hf_hub(name=model_name, device=torch_device)
    orig_model.eval()

    check_params_buffers(orig_model, torch_device)

    hf_model = transformers.AutoModel.from_pretrained(model_name)
    hf_model.to(torch_device)
    hf_model.eval()

    X_jit = torch.randint(0, hf_model.config.vocab_size, (3, 5), device=torch_device)
    mask_jit = torch.ones_like(X_jit, dtype=torch.bool)
    model, output = jit_method.convert(
        orig_model, with_torch_sdp, X_jit, AttentionMask(mask_jit)
    )

    torch.manual_seed(0)
    X = torch.randint(0, hf_model.config.vocab_size, (2, 10), device=torch_device)
    mask = torch.ones_like(X, dtype=torch.bool)

    with torch.no_grad():
        Y = output(model(X, AttentionMask(mask))).last_hidden_layer_state
        Y_hf = hf_model(X).last_hidden_state

    torch_assertclose(Y, Y_hf, atol=atol, rtol=rtol)

    assert_with_mask_output_equals_hf(
        orig_model, hf_model, torch_device, atol, rtol, jit_method
    )

    if check_config and jit_method == JITMethod.Disable:
        assert_model_config(model, Y)


def assert_decoder_with_cache_output_equals_hf(
    orig_model: DecoderModule,
    hf_model: "transformers.AutoModel",
    torch_device: torch.device,
    atol: float,
    rtol: float,
    jit_method: JITMethod,
    with_torch_sdp=False,
):
    X_jit = torch.randint(0, hf_model.config.vocab_size, (3, 5), device=torch_device)
    mask_jit = torch.ones_like(X_jit, dtype=torch.bool)
    with torch.no_grad():
        cache_jit = orig_model(X_jit, AttentionMask(mask_jit), store_cache=True).cache

    model, output = jit_method.convert(
        DecoderWithCache(orig_model),
        with_torch_sdp,
        X_jit,
        AttentionMask(torch.concat([mask_jit, mask_jit], dim=1)),
        cache_jit,
    )

    _, n_heads, _, head_width = cache_jit[0].key.shape
    empty_kv_jit = torch.zeros(
        (2, n_heads, 0, head_width),
        dtype=cache_jit[0].key.dtype,
        device=torch_device,
    )
    empty_cache_jit = [
        KeyValueCache(key=empty_kv_jit, value=empty_kv_jit)
    ] * hf_model.config.num_hidden_layers

    X = torch.randint(0, hf_model.config.vocab_size, (2, 10), device=torch_device)
    mask = torch.ones_like(X, dtype=torch.bool)
    X_rest = torch.randint(0, hf_model.config.vocab_size, (2, 10), device=torch_device)
    mask_rest = torch.cat([mask, torch.ones_like(X_rest, dtype=torch.bool)], dim=1)
    with torch.no_grad():
        Y = model(X, AttentionMask(mask), empty_cache_jit)
        Y_hf = hf_model(X, use_cache=True)
        Y = output(
            model(X_rest, AttentionMask(mask_rest), cache=output(Y).cache)
        ).last_hidden_layer_state
        Y_hf = hf_model(X_rest, past_key_values=Y_hf.past_key_values).last_hidden_state

    torch_assertclose(Y, Y_hf, atol=atol, rtol=rtol)


def assert_with_mask_output_equals_hf(
    orig_model: Union[DecoderModule, EncoderModule],
    hf_model: "transformers.AutoModel",
    torch_device: torch.device,
    atol: float,
    rtol: float,
    jit_method: JITMethod,
    with_torch_sdp=False,
):
    X_jit = torch.randint(0, hf_model.config.vocab_size, (3, 5), device=torch_device)
    mask_jit = torch.rand_like(X_jit, dtype=torch.float32) < 0.5
    model, output = jit_method.convert(
        orig_model, with_torch_sdp, X_jit, AttentionMask(mask_jit)
    )

    X = torch.randint(0, hf_model.config.vocab_size, (2, 10), device=torch_device)
    mask = torch.rand((2, 10), dtype=torch.float, device=torch_device) < 0.5
    with torch.no_grad():
        Y = output(
            model(X, AttentionMask(mask))
        ).last_hidden_layer_state * mask.unsqueeze(-1)
        Y_hf = hf_model(X, attention_mask=mask).last_hidden_state * mask.unsqueeze(-1)
    torch_assertclose(Y, Y_hf, atol=atol, rtol=rtol)


def assert_decoder_with_positions_equals_hf(
    orig_model: DecoderModule,
    hf_model: "transformers.AutoModel",
    torch_device: torch.device,
    atol: float,
    rtol: float,
    jit_method: JITMethod,
    with_torch_sdp=False,
):
    # NOTE: this test may break in the future because we are relying on a
    #       slicing bug, which results in more general support for positions
    #       (like our implementation):
    #
    #       https://github.com/huggingface/transformers/blob/f924df3c7e5150317ef47754a31cebf2893570ce/src/transformers/models/gpt_neox/modeling_gpt_neox.py#L278
    #
    #       However, it is still nice to test this with arbitrary positions. If
    #       this breaks, just replace max_position_embbeddings by the sequence
    #       length.

    X_jit = torch.randint(0, hf_model.config.vocab_size, (3, 5), device=torch_device)
    mask_jit = torch.ones_like(X_jit, dtype=torch.bool)
    positions_jit = torch.randint_like(X_jit, 10)
    model, output = jit_method.convert(
        DecoderWithPositions(orig_model),
        with_torch_sdp,
        X_jit,
        AttentionMask(mask_jit),
        positions_jit,
    )

    X = torch.randint(0, hf_model.config.vocab_size, (2, 10), device=torch_device)
    mask = torch.ones_like(X, dtype=torch.bool)
    positions = torch.randint(0, 10, (2, 10), device=torch_device)
    with torch.no_grad():
        Y = output(
            model(X, AttentionMask(mask), positions=positions)
        ).last_hidden_layer_state
        Y_hf = hf_model(X, position_ids=positions).last_hidden_state

    torch_assertclose(Y, Y_hf, atol=atol, rtol=rtol)


def assert_model_config(model: TransformerModule, model_output: Tensor):
    assert isinstance(model, TransformerModule)
    config = model.config

    hidden_width = model_output.size(-1)
    assert config.layer.feedforward.hidden_width == hidden_width


def assert_model_hf_serialization_roundtrip(
    model_class: Type[FromHF],
    model_name: str,
    torch_device: torch.device,
    *,
    model_revision: str = "main",
    atol: float = 1e-5,
    rtol: float = 1e-5,
    trust_remote_code: bool = False,
    optional_hf_config_keys: Set[str] = set(),
):
    orig_model = model_class.from_hf_hub(
        name=model_name,
        revision=model_revision,
        device=torch_device,
    )
    orig_model.eval()

    check_params_buffers(orig_model, torch_device)

    auto_cls = (
        transformers.AutoModelForCausalLM
        if isinstance(orig_model, CausalLMModule)
        else transformers.AutoModel
    )

    hf_model = auto_cls.from_pretrained(
        model_name,
        revision=model_revision,
        trust_remote_code=trust_remote_code,
    )
    hf_model.to(torch_device)
    hf_model.eval()

    hf_model_statedict = hf_model.state_dict()
    orig_model_hf_statedict = orig_model.state_dict_to_hf(orig_model.state_dict())
    for name in orig_model_hf_statedict.keys():
        assert name in hf_model_statedict.keys(), f"{name} not found in HF state dict"
        torch_assertclose(orig_model_hf_statedict[name], hf_model_statedict[name])

    hf_config = ModelRepository(
        HfHubRepository(name=model_name, revision=model_revision)
    ).model_config()
    orig_model_hf_config = orig_model.config_to_hf(orig_model.config)

    # These won't be the same anyway, so we can skip them.
    ignored_keys = ("architectures", "transformers_version")

    for k, v in orig_model_hf_config.items():
        if k in ignored_keys:
            continue
        elif k in optional_hf_config_keys and k not in hf_config:
            continue
        assert k in hf_config, f"Key '{k}' is missing in the Hugging Face model config"
        assert (
            hf_config[k] == v
        ), f"Key '{k}' value '{v}' is different in the Hugging Face model config ('{hf_config[k]}')"


def check_params_buffers(model: Module, device: torch.device):
    """
    Check that parameters/buffers are placed on the correct device and that
    parameters are leaf nodes.
    """
    for buffer in model.buffers():
        assert buffer.device == device

    for param in model.parameters():
        assert param.device == device
        assert param.is_leaf



import dataclasses
from typing import Any, Generic, List, Optional, Type, TypeVar

import torch

from ..models.auto_model import AutoCausalLM
from ..models.module import CausalLMModule
from ..models.output import CacheT
from ..quantization.bnb.config import BitsAndBytesConfig
from ..tokenizers.auto_tokenizer import AutoTokenizer
from ..tokenizers.chunks import InputChunks, TextChunk
from ..tokenizers.tokenizer import TokenizerBase
from .config import GeneratorConfig, SampleGeneratorConfig
from .generator import Generator
from .generator_wrapper import GeneratorWrapper
from .hf_hub import FromHF
from .string_generator import StringGenerator

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="DefaultGenerator")


class DefaultGenerator(Generic[CacheT], GeneratorWrapper, FromHF):
    """
    Generator wrapper for models that do not need specific prompting.
    """

    def __init__(
        self,
        tokenizer: TokenizerBase,
        causal_lm: CausalLMModule[Any, CacheT],
        default_config: Optional[GeneratorConfig] = None,
    ):
        """
        Construct a generic generator.

        :param tokenizer:
            A tokenizer.
        :param causal_lm:
            A causal language model.
        :param default_config:
            Configuration to use as a default when the configuration provided
            to the ``generate`` method is underspecified. For instance, if the
            end-of-sequence identifier is ``None`` in the generation
            configuration, it will be taken from the default configuration.
        """
        super().__init__()

        eos_id = (
            tokenizer.piece_to_id(tokenizer.eos_piece)
            if tokenizer.eos_piece is not None
            else None
        )

        self.generator = StringGenerator(tokenizer, Generator(causal_lm))
        self.default_config = (
            SampleGeneratorConfig(max_generated_pieces=200, eos_id=eos_id)
            if default_config is None
            else default_config
        )

    @classmethod
    def from_hf_hub_to_cache(
        cls: Type[Self],
        *,
        name: str,
        revision: str = "main",
    ):
        AutoTokenizer.from_hf_hub_to_cache(name=name, revision=revision)
        AutoCausalLM.from_hf_hub_to_cache(name=name, revision=revision)

    @classmethod
    def from_hf_hub(
        cls: Type[Self],
        *,
        name: str,
        revision: str = "main",
        device: Optional[torch.device] = None,
        quantization_config: Optional[BitsAndBytesConfig] = None,
    ) -> Self:
        tokenizer = AutoTokenizer.from_hf_hub(name=name, revision=revision)
        causal_lm = AutoCausalLM.from_hf_hub(
            name=name,
            revision=revision,
            device=device,
            quantization_config=quantization_config,
        )
        return cls(tokenizer, causal_lm)

    def generate(self, prompts: List[str], config: GeneratorConfig) -> List[str]:
        eos_id = self.default_config.eos_id if config.eos_id is None else config.eos_id
        max_generated_pieces = (
            self.default_config.max_generated_pieces
            if config.max_generated_pieces is None
            else config.max_generated_pieces
        )
        config = dataclasses.replace(
            config, eos_id=eos_id, max_generated_pieces=max_generated_pieces
        )
        return self.generator(self.preprocess_prompts(prompts), config=config)

    def preprocess_prompts(self, prompts: List[str]) -> List[InputChunks]:
        """
        Prepare a list of prompts for generation.

        :param prompts:
            The prompts to prepare.
        :returns:
            Prepared prompts.
        """
        return [InputChunks([TextChunk(prompt)]) for prompt in prompts]



import warnings
from abc import ABC, abstractmethod
from typing import Generic, Optional, Type, TypeVar

import torch
from catalogue import Registry
from fsspec import AbstractFileSystem

from curated_transformers import registry

from ..layers.cache import KeyValueCache
from ..quantization.bnb.config import BitsAndBytesConfig
from ..repository.fsspec import FsspecArgs, FsspecRepository
from ..repository.hf_hub import HfHubRepository
from ..repository.repository import ModelRepository, Repository
from .config import TransformerConfig
from .hf_hub import FromHF
from .module import CausalLMModule, DecoderModule, EncoderModule, TransformerModule

ModelT = TypeVar("ModelT")


class AutoModel(ABC, Generic[ModelT]):
    """
    Base class for models that can be loaded from the Hugging
    Face Model Hub.
    """

    _base_cls: Type[TransformerModule]
    _registry: Registry

    @classmethod
    def _resolve_model_cls(
        cls,
        repo: ModelRepository,
    ) -> Type[FromHF]:
        config = repo.model_config()

        for entrypoint, module_cls in cls._registry.get_entry_points().items():
            if not issubclass(module_cls, FromHF):
                warnings.warn(
                    f"Entry point `{entrypoint}` cannot load from Hugging Face Hub "
                    "since the FromHF mixin is not implemented"
                )
                continue

            if not issubclass(module_cls, cls._base_cls):
                warnings.warn(
                    f"Entry point `{entrypoint}` cannot be used by `{cls.__name__}` "
                    f"since it does does not have `{cls._base_cls.__name__}` "
                    "as its base class"
                )
                continue

            if module_cls.is_supported(config):
                return module_cls

        entrypoints = {
            entrypoint for entrypoint in cls._registry.get_entry_points().keys()
        }

        raise ValueError(
            f"Unsupported model type for `{cls.__name__}`. "
            f"Registered models: {', '.join(sorted(entrypoints))}"
        )

    @classmethod
    def _instantiate_model(
        cls,
        repo: Repository,
        device: Optional[torch.device],
        quantization_config: Optional[BitsAndBytesConfig],
    ) -> FromHF:
        module_cls = cls._resolve_model_cls(ModelRepository(repo))
        module = module_cls.from_repo(
            repo=repo,
            device=device,
            quantization_config=quantization_config,
        )
        return module

    @classmethod
    def from_fsspec(
        cls,
        *,
        fs: AbstractFileSystem,
        model_path: str,
        fsspec_args: Optional[FsspecArgs] = None,
        device: Optional[torch.device] = None,
        quantization_config: Optional[BitsAndBytesConfig] = None,
    ) -> ModelT:
        """
        Construct a module and load its parameters from a fsspec filesystem.

        :param fs:
            The filesystem to load the model from.
        :param model_path:
            The path of the model on the filesystem.
        :param fsspec_args:
            Implementation-specific keyword arguments to pass to fsspec
            filesystem operations.
        :param device:
            Device on which the model is initialized.
        :param quantization_config:
            Configuration for loading quantized weights.
        :returns:
            Module with the parameters loaded.
        """
        return cls.from_repo(
            repo=FsspecRepository(
                fs,
                path=model_path,
                fsspec_args=fsspec_args,
            ),
            device=device,
            quantization_config=quantization_config,
        )

    @classmethod
    def from_hf_hub(
        cls,
        *,
        name: str,
        revision: str = "main",
        device: Optional[torch.device] = None,
        quantization_config: Optional[BitsAndBytesConfig] = None,
    ) -> ModelT:
        """
        Construct and load a model or a generator from Hugging Face Hub.

        :param name:
            Model name.
        :param revision:
            Model revision.
        :param device:
            Device on which to initialize the model.
        :param quantization_config:
            Configuration for loading quantized weights.
        :returns:
            Loaded model or generator.
        """
        return cls.from_repo(
            repo=HfHubRepository(name=name, revision=revision),
            device=device,
            quantization_config=quantization_config,
        )

    @classmethod
    @abstractmethod
    def from_repo(
        cls,
        *,
        repo: Repository,
        device: Optional[torch.device] = None,
        quantization_config: Optional[BitsAndBytesConfig] = None,
    ) -> ModelT:
        """
        Construct and load a model or a generator from a repository.

        :param repository:
            The repository to load from.
        :param device:
            Device on which to initialize the model.
        :param quantization_config:
            Configuration for loading quantized weights.
        :returns:
            Loaded model or generator.
        """

        raise NotImplementedError

    @classmethod
    def from_hf_hub_to_cache(
        cls,
        *,
        name: str,
        revision: str = "main",
    ):
        """
        Download the model's weights from Hugging Face Hub into the local
        Hugging Face cache directory. Subsequent loading of the
        model will read the weights from disk. If the weights are already
        cached, this is a no-op.

        :param name:
            Model name.
        :param revision:
            Model revision.
        """
        repo = ModelRepository(HfHubRepository(name=name, revision=revision))
        repo.model_config()
        repo.model_checkpoints()


class AutoEncoder(AutoModel[EncoderModule[TransformerConfig]]):
    """
    Encoder model loaded from the Hugging Face Model Hub.
    """

    _base_cls = EncoderModule
    _registry: Registry = registry.encoders

    @classmethod
    def from_repo(
        cls,
        *,
        repo: Repository,
        device: Optional[torch.device] = None,
        quantization_config: Optional[BitsAndBytesConfig] = None,
    ) -> EncoderModule[TransformerConfig]:
        encoder = cls._instantiate_model(repo, device, quantization_config)
        assert isinstance(encoder, EncoderModule)
        return encoder


class AutoDecoder(AutoModel[DecoderModule[TransformerConfig, KeyValueCache]]):
    """
    Decoder module loaded from the Hugging Face Model Hub.
    """

    _base_cls = DecoderModule
    _registry = registry.decoders

    @classmethod
    def from_repo(
        cls,
        *,
        repo: Repository,
        device: Optional[torch.device] = None,
        quantization_config: Optional[BitsAndBytesConfig] = None,
    ) -> DecoderModule[TransformerConfig, KeyValueCache]:
        decoder = cls._instantiate_model(repo, device, quantization_config)
        assert isinstance(decoder, DecoderModule)
        return decoder


class AutoCausalLM(AutoModel[CausalLMModule[TransformerConfig, KeyValueCache]]):
    """
    Causal LM model loaded from the Hugging Face Model Hub.
    """

    _base_cls = CausalLMModule
    _registry: Registry = registry.causal_lms

    @classmethod
    def from_repo(
        cls,
        *,
        repo: Repository,
        device: Optional[torch.device] = None,
        quantization_config: Optional[BitsAndBytesConfig] = None,
    ) -> CausalLMModule[TransformerConfig, KeyValueCache]:
        causal_lm = cls._instantiate_model(repo, device, quantization_config)
        assert isinstance(causal_lm, CausalLMModule)
        return causal_lm



import pytest
import torch

from curated_transformers.generation.config import (
    GreedyGeneratorConfig,
    SampleGeneratorConfig,
)
from curated_transformers.generation.dolly_v2 import DollyV2Generator

from ..conftest import GPU_TESTS_ENABLED


@pytest.fixture(scope="module")
def dolly_generator():
    return DollyV2Generator.from_hf_hub(
        name="databricks/dolly-v2-3b", device=torch.device("cuda", index=0)
    )


@pytest.mark.slow
@pytest.mark.skipif(not GPU_TESTS_ENABLED, reason="requires GPU")
def test_generate_deterministic(dolly_generator):
    prompts = [
        "What is the Rust programming language?",
        "What is spaCy?",
    ]
    answers = [
        "Rust is a multi-paradigm, high-level, general-purpose programming language. Rust is designed to have a small, consistent, and predictable language surface. Rust is also designed to be efficient, and to have a small memory footprint. Rust is designed to be safe, and to have a well-defined memory model. Rust is also designed to be concurrent, and to have a good support for concurrent programming. Rust is designed to be fast, and to have a good support for performance-critical code. Rust is also designed to be modular, and to have a good support for modular programming. Rust is designed to have a good support for internationalization and localization.\n\n",
        "SpaCy is a natural language processing (NLP) library for Python that provides tokenization, part-of-speech (POS) tagging, named entity recognition (NER), and dependency parsing.\n\n",
    ]
    assert dolly_generator(prompts, config=GreedyGeneratorConfig()) == answers

    # Test in reverse order to verify that sequence identifiers are
    # handled correctly.
    assert (
        dolly_generator(prompts[::-1], config=GreedyGeneratorConfig()) == answers[::-1]
    )


@pytest.mark.slow
@pytest.mark.skipif(not GPU_TESTS_ENABLED, reason="requires GPU")
def test_generate_max_generated_pieces(dolly_generator):
    prompts = [
        "What is the Rust programming language?",
        "What is spaCy?",
    ]
    answers = ["Rust is a multi-paradigm,", "SpaCy is a natural language processing (N"]

    assert (
        dolly_generator(prompts, config=GreedyGeneratorConfig(max_generated_pieces=10))
        == answers
    )

    # Test in reverse order to verify that sequence identifiers are
    # handled correctly.
    assert (
        dolly_generator(
            prompts[::-1], config=GreedyGeneratorConfig(max_generated_pieces=10)
        )
        == answers[::-1]
    )


@pytest.mark.slow
@pytest.mark.skipif(not GPU_TESTS_ENABLED, reason="requires GPU")
def test_generate_sample(dolly_generator):
    prompts = [
        "What is spaCy?",
        "What is spaCy?",
    ]
    # Fix the seed so that we are always randomly sampling in the same way.
    torch.manual_seed(0)
    assert dolly_generator(prompts, config=SampleGeneratorConfig(top_k=10)) == [
        "SpaCy (short for spaCy Toolkit) is a Python library for Natural Language Processing (NLP) and machine translation based on the   spaCy research project. It has been open-source since June 2023 and has been used in production for NLP tasks like POS tag classification (part of OpenStax' TACL), question answering (QACQUEL), and semantic parsing (CoreNLP) as of October 2023.\n\n",
        "SpaCy is an open-source package for Python that automates part of the process of building language models (LMs), or parsers for plain text, using a technique called graph-based learning. It supports English, French, German, Dutch, Italian, Spanish, Portuguese, Polish, and Dutch, all of which are official language groups of Europe.\n\n",
    ]

    torch.manual_seed(0)
    assert dolly_generator(
        prompts, config=SampleGeneratorConfig(top_k=5, temperature=2)
    ) == [
        "SpaCy (short for Spanish Language Model) is a natural language processor (NLP) for English based on the open source spaCy project. It supports sentence detection, part-of-speech tagging and coreference chains, among other things. spaCy was originally developed at the University of Cambridge's Machine Translation Lab. Since then, it has also developed a version optimized for Python that can be installed viapip.\n\n",
        "The spaCy library is an NLP library based on Stanford's PoSpell architecture, that is designed to make NLP easy. spaCy supports all the standard NLTK pipeline stages and is able to outperform the NLTK on many Named Entity Tagging tasks, as well many others, both at test-level and at the system-wide average. It has a similar architecture to NLTK’s, but is designed from the ground up for the needs of a NLP research community rather than a production system: It has a smaller and less mature API, does away with its core tokenizer (which is notoriously hard to train), and is based on the Speller system from Stanford's CS-ADLDN program, which has been shown to significantly outperform the NLTK tokeniser in terms of both accuracy (93.7% on a test set of 20K tokens vs. NLTK's 76.3%) and efficiency.\n\n",
    ]



import re
from abc import ABC, abstractmethod
from typing import Optional, Tuple


class StringTransform(ABC):
    """
    Base class for reversible string transformations.
    """

    def __init__(self, reversible: bool = True):
        super().__init__()
        self._reversible = reversible

    @abstractmethod
    def _apply(self, string: str) -> str:
        raise NotImplementedError

    @abstractmethod
    def _revert(self, string: str) -> str:
        raise NotImplementedError

    def apply(self, string: str) -> str:
        """
        Applies the transformation to the given string.

        :param string:
            String to transform.
        :returns:
            Transformed string.
        """
        return self._apply(string)

    def revert(self, string: str) -> str:
        """
        Reverts the previously applied transformation of the given string.

        :param string:
            Previously transformed string.
        :returns:
            Reverted string.
        """
        if self._reversible:
            return self._revert(string)
        else:
            return string


class StringTransformations:
    """
    Provides factory methods for different string transformations.
    """

    @staticmethod
    def regex_sub(
        forward: Tuple[str, str], backward: Optional[Tuple[str, str]]
    ) -> StringTransform:
        """
        Factory method to construct a string substitution transform
        using regular expressions.

        :param forward:
            Tuple where the first string is a RegEx pattern
            and the second the replacement.

            This operation is performed when the :meth:`.apply`
            method is invoked.
        :param backward:
            Optional tuple where the first string is a RegEx pattern
            and the second the replacement.

            This operation is performed when the :meth:`.revert`
            method is invoked. If ``None``, it is a no-op.
        """
        return StringSubRegEx(forward, backward)

    @staticmethod
    def sub(
        substring: str, replacement: str, *, reversible: bool = True
    ) -> StringTransform:
        """
        Factory method to construct a string substitution transform.

        :param substring:
            The substring to be replaced.
        :param replacement:
            The replacement string.
        :param reversible:
            If the reverse transformation is to
            be performed.
        """
        return StringSub(substring, replacement, reversible=reversible)

    @staticmethod
    def replace(
        replacee: str, replacement: str, *, reversible: bool = True
    ) -> StringTransform:
        """
        Factory method to construct a full string replacement transform.

        :param replacee:
            The full string to be replaced.
        :param replacement:
            The replacement string.
        :param reversible:
            If the reverse transformation is to
            be performed.
        """
        return StringReplace(replacee, replacement, reversible=reversible)

    @staticmethod
    def remove_prefix(prefix: str, *, reversible: bool = True) -> StringTransform:
        """
        Factory method to construct a string prefix removal transform.

        :param prefix:
            Prefix to be removed.
        :param reversible:
            If the reverse transformation is to
            be performed.
        """
        return StringRemovePrefix(prefix, reversible=reversible)


class StringSubRegEx(StringTransform):
    """
    Substitute a substring with another string using
    regular expressions.
    """

    def __init__(self, forward: Tuple[str, str], backward: Optional[Tuple[str, str]]):
        """
        Construct a reversible substitution.

        :param forward:
            Tuple where the first string is a RegEx pattern
            and the second the replacement.

            This operation is performed when the :meth:`.apply`
            method is invoked.
        :param backward:
            Optional tuple where the first string is a RegEx pattern
            and the second the replacement.

            This operation is performed when the :meth:`.revert`
            method is invoked. If ``None``, it is a no-op.
        """
        super().__init__(backward is not None)

        self.forward = forward
        self.backward = backward

    def _apply(self, string: str) -> str:
        return re.sub(self.forward[0], self.forward[1], string)

    def _revert(self, string: str) -> str:
        if self.backward is None:
            raise ValueError("Attempting to revert an irreversible string transform")
        return re.sub(self.backward[0], self.backward[1], string)


class StringSub(StringTransform):
    """
    Substitute a substring with another string.
    """

    def __init__(self, substring: str, replacement: str, *, reversible: bool = True):
        """
        Construct a reversible substitution.

        :param substring:
            The substring to be replaced.
        :param replacement:
            The replacement string.
        :param reversible:
            If the reverse transformation is to
            be performed.
        """
        super().__init__(reversible)
        self.substring = substring
        self.replacement = replacement

    def _apply(self, string: str) -> str:
        return string.replace(self.substring, self.replacement)

    def _revert(self, string: str) -> str:
        return string.replace(self.replacement, self.substring)


class StringReplace(StringTransform):
    """
    Replaces an entire string with another.
    """

    def __init__(self, replacee: str, replacement: str, *, reversible: bool = True):
        """
        Construct a reversible replacement.

        :param replacee:
            The full string to be replaced.
        :param replacement:
            The replacement string.
        :param reversible:
            If the reverse transformation is to
            be performed.
        """
        super().__init__(reversible)
        self.replacee = replacee
        self.replacement = replacement

    def _apply(self, string: str) -> str:
        if string == self.replacee:
            return self.replacement
        else:
            return string

    def _revert(self, string: str) -> str:
        if string == self.replacement:
            return self.replacee
        else:
            return string


class StringRemovePrefix(StringTransform):
    """
    Strips a prefix from a given string.
    """

    def __init__(self, prefix: str, *, reversible: bool = True):
        """
        Construct a reversible left strip.

        :param prefix:
            Prefix to be removed.
        :param reversible:
            If the reverse transformation is to
            be performed.
        """

        super().__init__(reversible)
        self.prefix = prefix

    def _apply(self, string: str) -> str:
        # TODO: Should be replaced with `removeprefix` once
        # Python 3.9 is the minimum requirement.
        return re.sub(f"^{re.escape(self.prefix)}", "", string)

    def _revert(self, string: str) -> str:
        return f"{self.prefix}{string}"



from copy import deepcopy

from curated_transformers.tokenizers.chunks import (
    InputChunks,
    MergedSpecialPieceChunk,
    SpecialPieceChunk,
    TextChunk,
)


def test_text_chunk_merge():
    chunks = InputChunks(
        [
            TextChunk(" "),
            TextChunk("Hello world!"),
            TextChunk(" "),
        ]
    )

    # Ensure that we are not doing accidental modification to the
    # object itself.
    chunks_copy = deepcopy(chunks)
    chunks.merge_text_chunks()
    assert chunks == chunks_copy

    assert chunks.merge_text_chunks() == [
        TextChunk(" Hello world! "),
    ]


def test_special_text_special_chunk_merge():
    chunks = InputChunks(
        [
            SpecialPieceChunk("<s>", after=" "),
            TextChunk("Hello world!"),
            SpecialPieceChunk("</s>", before=" "),
        ]
    )

    # Ensure that we are not doing accidental modification to the
    # object itself.
    chunks_copy = deepcopy(chunks)
    chunks.merge_text_chunks()
    assert chunks == chunks_copy

    assert chunks.merge_text_chunks() == [
        MergedSpecialPieceChunk("<s>"),
        TextChunk(" Hello world! "),
        MergedSpecialPieceChunk("</s>"),
    ]


def test_special_special_chunk_merge():
    chunks = InputChunks(
        [
            SpecialPieceChunk("<s>", after=" "),
            SpecialPieceChunk("</s>", before=" "),
        ]
    )

    # Ensure that we are not doing accidental modification to the
    # object itself.
    chunks_copy = deepcopy(chunks)
    chunks.merge_text_chunks()
    assert chunks == chunks_copy

    assert chunks.merge_text_chunks() == [
        MergedSpecialPieceChunk("<s>"),
        TextChunk("  "),
        MergedSpecialPieceChunk("</s>"),
    ]



from dataclasses import dataclass
from typing import Optional

import torch

from ..layers.activations import Activation


@dataclass
class RotaryEmbeddingConfig:
    """
    Configuration options for rotary embeddings (`Su et al., 2021`_).

    .. _Su et al., 2021: https://arxiv.org/abs/2104.09864

    :param rotary_base:
        Base in signifying the rotary embedding period.
    :param rotary_fraction:
        Fraction of hidden width to apply rotary embeddings to.
        Must be in ``[0,1]``.
    """

    rotary_base: int
    rotary_fraction: float


@dataclass
class TransformerAttentionLayerConfig:
    """
    Configuration options for self-attention.

    :param dropout_prob:
        Dropout probabilty to apply after attention.
    :param hidden_width:
        Hidden width of the transformer.
    :param n_query_heads:
        Number of attention heads.
    :param n_key_value_heads:
        Number of key and value heads.
    :param rotary_embeddings:
        Rotary embedding configuration.
    :param use_alibi:
        Use ALiBi linear biases.
    :param use_bias:
        Use bias in linear layers.
    :param use_parallel_attention:
        Use parallel attention.
    """

    dropout_prob: float
    hidden_width: int
    n_query_heads: int
    n_key_value_heads: int
    rotary_embeddings: Optional[RotaryEmbeddingConfig]
    use_alibi: bool
    use_bias: bool
    use_parallel_attention: bool


@dataclass
class TransformerEmbeddingLayerConfig:
    """
    Configuration options for embeddings.

    :param dropout_prob:
        Dropout probabilty for the embedding layer.
    :param embedding_width:
        Width of the embedding representations.
    :param layer_norm_eps:
        Epsilon for layer normalization.
    :param n_positions:
        Maximum length of position embeddings.
    :param n_pieces:
        Vocabulary size (number of embeddings).
    :param n_types:
        Token type vocabulary size (number of token type embeddings).
    """

    dropout_prob: float
    embedding_width: int
    layer_norm_eps: float
    n_positions: Optional[int]
    n_pieces: int
    n_types: Optional[int]


@dataclass
class TransformerFeedForwardLayerConfig:
    """
    Configuration options for transformer feed-forward layers.

    :param activation:
        Activation in the feed-forward layer
    :param hidden_width:
        Hidden width of the transformer.
    :param intermediate_width:
        Intermediate width in the feed-forward layer.
    :param use_bias:
        Use bias in linear layers.
    :param use_gate:
        Use Gated Linear Units.
    """

    activation: Activation
    hidden_width: int
    intermediate_width: int
    use_bias: bool
    use_gate: bool


@dataclass
class TransformerLayerConfig:
    """
    Configuration options for transformer layers.

    :param attention:
        Attention layer config.
    :param dropout_prob:
        Dropout probabilty to apply after hidden layers.
    :param feedforward:
        Feed-forward layer config.
    :param layer_norm_eps:
        Epsilon for layer normalization.
    :param n_hidden_layers:
        Number of hidden layers.
    """

    attention: TransformerAttentionLayerConfig
    dropout_prob: float
    feedforward: TransformerFeedForwardLayerConfig
    layer_norm_eps: float
    n_hidden_layers: int


@dataclass
class TransformerConfig:
    """
    Configuration options for a transformer model.

    :param embedding:
        Embedding layer config.
    :param layer:
        Transformer hidden layer config.
    :param dtype:
        Default data type used by the model's
        parameters.
    """

    embedding: TransformerEmbeddingLayerConfig
    layer: TransformerLayerConfig
    dtype: torch.dtype



import math
from typing import Optional, Tuple

import torch
from torch import Tensor
from torch.nn import Module

from .cache import KeyValueCache


# https://pytorch.org/tutorials/beginner/transformer_tutorial.html
class SinusoidalPositionalEmbedding(Module):
    """
    Sinusoidal positional embeddings (`Vaswani et al., 2017`_).

    .. _Vaswani et al., 2017: https://arxiv.org/abs/1706.03762
    """

    def __init__(
        self,
        *,
        width: int,
        max_len: int,
        normalize=True,
        device: Optional[torch.device] = None,
    ):
        """
        Construct a sinusoidal positional embedding module.

        :param width:
            Width of the embedding.
        :param max_len:
            Maximum length of the embedding.
        :param normalize:
            Perform L2 normalization of the embedding.
        :param device:
            Device on which the module is to be initialized.
        """
        super().__init__()

        position = torch.arange(max_len, device=device).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, width, 2, device=device) * (-math.log(10000.0) / width)
        )

        pe = torch.zeros(max_len, width, device=device)
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)

        if normalize == True:
            l2 = torch.linalg.vector_norm(pe, dim=-1)
            pe /= l2.unsqueeze(-1)

        self.pe = pe
        self.pe.requires_grad = False

    def forward(self, input: Tensor) -> Tensor:
        """
        Returns the positional embedding for the input.

        :param input:
            Input tensor.

            *Shape:* ``(batch_size, seq_len)``
        :returns:
            Positional embedding for the input.

            *Shape:* ``(seq_len, width)``
        """
        return self.pe[: input.size(1), :]


class RotaryEmbeddings(Module):
    """
    Rotary embeddings (`Su et al., 2021`_).

    .. _Su et al., 2021: https://arxiv.org/abs/2104.09864
    """

    cos: Tensor
    sin: Tensor
    theta: Tensor

    def __init__(
        self,
        width: int,
        *,
        seq_len: int = 512,
        base: int = 10000,
        device: Optional[torch.device] = None,
    ):
        """
        Construct a rotary embedding module. The rotary embedding
        will be precomputed for up to ``seq_len`` positions. The embedding
        will be recomputed when a longer sequence is found in the input.

        :param width:
            Rotary embedding width.
            Must be even.
        :param seq_len:
            Number of positions to initially precompute.
        :param base:
            The base used for :math:`\\theta_i`.
            Determines the cycle length of the embeddings.
        :param device:
            Device on which the module is to be initialized.
        """
        super().__init__()

        if width % 2:
            raise ValueError(f"Width of rotary embeddings must be even, was: {width}")

        # Ignore allocations on the meta device as we don't persist our buffer,
        # i.e., we don't expect the backing tensor to be replaced with pretrained weights.
        if device is not None and device.type == "meta":
            device = None
        # Θ_i = 10000^(-2(i-1)/d)
        theta = torch.pow(
            base, -torch.arange(0, width, 2, dtype=torch.float, device=device) / width
        )
        self.register_buffer("theta", theta, persistent=False)

        self._create_rotary_embed(width=width, length=seq_len)

    def _create_rotary_embed(self, *, width: int, length: int):
        # mΘ
        position = torch.arange(length, device=self.theta.device).unsqueeze(1)
        m_theta = position * self.theta.unsqueeze(0)

        # We apply both sin and cos twice (see Eq 15, 34), but the ordering
        # is changed for compatibility with most common implementations.
        m_theta = torch.cat([m_theta, m_theta], dim=-1)

        re_cos = m_theta.cos().view([length, width])
        re_sin = m_theta.sin().view([length, width])

        self.register_buffer("cos", re_cos, persistent=False)
        self.register_buffer("sin", re_sin, persistent=False)

    def _rotate(self, input: Tensor):
        """
        Rotate the input tensor by half of its innermost width.

        :param input:
            Tensor to rotate.

            *Shape:* ``(..., width)``
        :returns:
            Rotated tensor.

            *Shape:* ``(.., width)``

        :meta private:
        """
        half_idx = input.shape[-1] // 2
        input_1 = -input[..., half_idx:]
        input_2 = input[..., :half_idx]
        return torch.cat([input_1, input_2], dim=-1)

    def forward(self, input: torch.Tensor, *, positions: Optional[Tensor] = None):
        """
        Apply rotary embeddings to the input.

        :param input:
            Input to apply the rotary embeddings to.

            *Shape:* ``(batch_size, n_heads, seq_len, width_per_head)``
        :param positions:
            Positions of the inputs. If no positions are
            provided, they are assumed to be ``[0, seq_len)``.

            *Shape:* ``(batch_size, seq_len)``
        :returns:
            Input with the rotary embeddings applied.

            *Shape:* ``(batch_size, n_heads, seq_len, width_per_head)``
        """
        batch_size, _, seq_len, width = input.shape

        if positions is None:
            # Fastpath: positions from [0..seq_len), avoid indexing.
            if self.cos.size(-2) < seq_len:
                self._create_rotary_embed(width=width, length=seq_len)
            rot_cos = self.cos[:seq_len, :].view(1, 1, seq_len, width)
            rot_sin = self.sin[:seq_len, :].view(1, 1, seq_len, width)
        else:
            max_len = int(positions.max()) + 1
            if self.cos.size(-2) < max_len:
                self._create_rotary_embed(width=width, length=max_len)

            # Flatten positions to index cos/sin arrays, then unflatten.
            #
            # Example shapes:
            #
            #   positions_flat - (batch_size * seq_len)
            #   self.cos - (max_len, width)
            #   rot_cos - (batch_size, seq_len, width)
            positions_flat = positions.view(-1)
            rot_cos = self.cos[positions_flat].view(batch_size, 1, seq_len, width)
            rot_sin = self.sin[positions_flat].view(batch_size, 1, seq_len, width)

        # Eq 34 with ordering changed for compatibility.
        return rot_cos * input + rot_sin * self._rotate(input)


class QueryKeyRotaryEmbeddings(Module):
    """
    Rotary embeddings (`Su et al., 2021`_) applied to query
    and key representations.

    .. _Su et al., 2021: https://arxiv.org/abs/2104.09864
    """

    def __init__(
        self,
        *,
        base: int = 10000,
        fraction: float,
        head_width: int,
        device: Optional[torch.device] = None,
    ) -> None:
        """
        Construct a rotary embedding module.

        :param base:
            Base in signifying the rotary embedding period.
        :param fraction:
            Fraction of hidden width to apply rotary embeddings to.
            Must be in ``[0,1]``.
        :param head_width:
            Width of key and value heads.
        :param device:
            Device on which the module is to be initialized.
        """
        super().__init__()
        if not 0.0 <= fraction <= 1.0:
            raise ValueError(
                f"Rotary embedding fraction should be between 0.0 and 1.0 inclusive, was: {fraction}"
            )
        self.rotary_width = int(fraction * head_width)
        self.head_width = head_width
        self.rotary_embeds = RotaryEmbeddings(
            width=self.rotary_width, base=base, device=device
        )

    def forward(
        self,
        *,
        query: Tensor,
        key: Tensor,
        cache: Optional[KeyValueCache] = None,
        positions: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Tensor]:
        """
        Apply rotary embeddings to the query and key.

        :param query:
            Query representations.

            *Shape:* ``(batch_size, head, seq_len, width_per_head)``
        :param key:
            Key representations.

            *Shape:* ``(batch_size, head, seq_len, width_per_head)``
        :param cache: Key/value cache to avoid recomputing
            key/value representations for tokens that were previously seen.
        :param positions: Input positions. Positions are needed to
            look up rotary embeddings. Normally, these positions are calculated
            automatically. But if the positions deviate for some reason, they
            can be provided through this argument.

            *Shape:* ``(batch_size, seq_len)``
        :returns:
            Query and key with the rotary embeddings applied.

            *Shape:* ``(batch_size, head, seq_len, width_per_head)``
        """
        head_width = self.head_width
        rotary_width = self.rotary_width

        # If a cache was provided, but no positions, assume that the
        # positions of the current batch continue from the cache.
        if cache is not None and positions is None:
            cache_len = cache.key.size(-2)
            seq_len = key.size(-2)
            positions = torch.arange(
                cache_len,
                cache_len + seq_len,
                dtype=torch.long,  # `torch.int32` isn't supported in indexing operations prior in torch<2.0.0.
                device=key.device,
            ).repeat(key.size(0), 1)

        if rotary_width == head_width:
            # Fast path: we apply rotary embeddings the full key/query vectors.
            key = self.rotary_embeds(key, positions=positions)
            query = self.rotary_embeds(query, positions=positions)
        else:
            # Otherwise, split up key/query vectors, apply rotary embeddings
            # and concatenate again.
            k_rotary, k_rest = key.split([rotary_width, head_width - rotary_width], -1)
            q_rotary, q_rest = query.split(
                [rotary_width, head_width - rotary_width], -1
            )

            # Apply rotary embeddings.
            k_rotary = self.rotary_embeds(k_rotary, positions=positions)
            q_rotary = self.rotary_embeds(q_rotary, positions=positions)

            query = torch.cat([q_rotary, q_rest], dim=-1)
            key = torch.cat([k_rotary, k_rest], dim=-1)

        return query, key



import pytest

from curated_transformers.models.hf_hub.conversion import CommonHFKeys
from curated_transformers.models.llama._hf import HFConfigKeys
from curated_transformers.models.llama.causal_lm import LlamaCausalLM

from ...compat import has_hf_transformers, has_torch_compile
from ...conftest import TORCH_DEVICES
from ..util import (
    JITMethod,
    assert_causal_lm_output_equals_hf,
    assert_model_hf_serialization_roundtrip,
)

LLAMA_TEST_MODELS = [
    "trl-internal-testing/tiny-random-LlamaForCausalLM",
    "explosion-testing/llama2-fewer-kv-heads",
    "explosion-testing/llama2-kv-sharing",
]


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
@pytest.mark.parametrize("with_torch_sdp", [False, True])
@pytest.mark.parametrize("model", LLAMA_TEST_MODELS)
def test_causal_lm(torch_device, model, with_torch_sdp):
    assert_causal_lm_output_equals_hf(
        LlamaCausalLM,
        model,
        torch_device,
        with_torch_sdp=with_torch_sdp,
    )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.skipif(not has_torch_compile, reason="requires torch.compile")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
@pytest.mark.parametrize("model", LLAMA_TEST_MODELS)
@pytest.mark.parametrize("with_torch_sdp", [False, True])
def test_causal_lm_torch_compile(torch_device, model, with_torch_sdp):
    assert_causal_lm_output_equals_hf(
        LlamaCausalLM,
        model,
        torch_device,
        jit_method=JITMethod.TorchCompile,
        with_torch_sdp=with_torch_sdp,
    )


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("model", LLAMA_TEST_MODELS)
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
def test_causal_lm_hf_serializtion_roundtrip(model, torch_device):
    assert_model_hf_serialization_roundtrip(
        LlamaCausalLM,
        model,
        torch_device,
        optional_hf_config_keys={
            HFConfigKeys.NUM_KEY_VALUE_HEADS.name,
            CommonHFKeys.ATTENTION_PROBS_DROPOUT_PROB.name,
            CommonHFKeys.HIDDEN_DROPOUT_PROB.name,
        },
    )



from typing import Any, Dict, Mapping, Optional

import torch

from ..roberta.config import RoBERTaConfig
from ..roberta.encoder import RoBERTaEncoder
from ._hf import _config_from_hf, _config_to_hf


class XLMREncoder(RoBERTaEncoder):
    """
    XLM-RoBERTa (`Conneau et al., 2019`_) encoder.

    .. _Conneau et al., 2019: https://arxiv.org/abs/1911.02116
    """

    def __init__(self, config: RoBERTaConfig, *, device: Optional[torch.device] = None):
        """
        Construct a XLM-RoBERTa encoder.

        :param config:
            Encoder configuration.
        :param device:
            Device to which the module is to be moved.
        :returns:
            The encoder.
        """
        super().__init__(config, device=device)

    @classmethod
    def is_supported(cls, config: Dict[str, Any]) -> bool:
        return config.get("model_type") == "xlm-roberta"

    @classmethod
    def config_from_hf(cls, hf_config: Mapping[str, Any]) -> RoBERTaConfig:
        return _config_from_hf(hf_config)

    @classmethod
    def config_to_hf(cls, curated_config: RoBERTaConfig) -> Mapping[str, Any]:
        return _config_to_hf(curated_config)



import pytest
import torch

from curated_transformers.tokenizers import PiecesWithIds
from curated_transformers.tokenizers.legacy import RoBERTaTokenizer

from ...compat import has_hf_transformers
from ...utils import torch_assertclose
from ..util import compare_tokenizer_outputs_with_hf_tokenizer


@pytest.fixture
def toy_tokenizer_from_files(test_dir):
    return RoBERTaTokenizer.from_files(
        vocab_file=test_dir / "toy-vocab.json", merges_file=test_dir / "toy-merges.txt"
    )


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_from_hf_hub_equals_hf_tokenizer(sample_texts):
    compare_tokenizer_outputs_with_hf_tokenizer(
        sample_texts, "roberta-base", RoBERTaTokenizer
    )


def test_from_files(toy_tokenizer_from_files, short_sample_texts):
    encoding = toy_tokenizer_from_files(short_sample_texts)
    _check_toy_tokenizer(encoding)


def _check_toy_tokenizer(pieces):
    assert isinstance(pieces, PiecesWithIds)
    assert len(pieces.ids) == 3
    assert len(pieces.pieces) == 3

    assert pieces.ids == [
        [0, 44, 997, 262, 305, 334, 79, 342, 262, 388, 79, 302, 70, 472, 72, 17, 2],
        [
            0,
            55,
            841,
            321,
            362,
            579,
            324,
            294,
            291,
            494,
            131,
            106,
            270,
            307,
            79,
            15,
            298,
            303,
            86,
            287,
            317,
            4,
            2,
        ],
        [
            0,
            55,
            494,
            283,
            86,
            513,
            468,
            506,
            78,
            81,
            428,
            297,
            163,
            229,
            252,
            697,
            346,
            163,
            227,
            104,
            298,
            294,
            264,
            163,
            227,
            238,
            490,
            83,
            276,
            69,
            323,
            169,
            252,
            253,
            884,
            70,
            772,
            480,
            295,
            911,
            17,
            2,
        ],
    ]

    assert pieces.pieces == [
        [
            "<s>",
            "I",
            "Ġsaw",
            "Ġa",
            "Ġg",
            "ir",
            "l",
            "Ġwith",
            "Ġa",
            "Ġte",
            "l",
            "es",
            "c",
            "op",
            "e",
            ".",
            "</s>",
        ],
        [
            "<s>",
            "T",
            "od",
            "ay",
            "Ġwe",
            "Ġwill",
            "Ġe",
            "at",
            "Ġp",
            "ok",
            "Ã",
            "©",
            "Ġb",
            "ow",
            "l",
            ",",
            "Ġl",
            "ot",
            "s",
            "Ġof",
            "Ġit",
            "!",
            "</s>",
        ],
        [
            "<s>",
            "T",
            "ok",
            "en",
            "s",
            "Ġwhich",
            "Ġare",
            "Ġun",
            "k",
            "n",
            "own",
            "Ġin",
            "ã",
            "ĥ",
            "ļ",
            "Ġmo",
            "st",
            "ã",
            "ģ",
            "§",
            "Ġl",
            "at",
            "in",
            "ã",
            "ģ",
            "Į",
            "Ġal",
            "p",
            "ha",
            "b",
            "et",
            "é",
            "ļ",
            "Ľ",
            "Ġvo",
            "c",
            "ab",
            "ul",
            "ar",
            "ies",
            ".",
            "</s>",
        ],
    ]

    torch_assertclose(
        pieces.padded_tensor(padding_id=1),
        torch.tensor(
            [
                [
                    0,
                    44,
                    997,
                    262,
                    305,
                    334,
                    79,
                    342,
                    262,
                    388,
                    79,
                    302,
                    70,
                    472,
                    72,
                    17,
                    2,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                ],
                [
                    0,
                    55,
                    841,
                    321,
                    362,
                    579,
                    324,
                    294,
                    291,
                    494,
                    131,
                    106,
                    270,
                    307,
                    79,
                    15,
                    298,
                    303,
                    86,
                    287,
                    317,
                    4,
                    2,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                ],
                [
                    0,
                    55,
                    494,
                    283,
                    86,
                    513,
                    468,
                    506,
                    78,
                    81,
                    428,
                    297,
                    163,
                    229,
                    252,
                    697,
                    346,
                    163,
                    227,
                    104,
                    298,
                    294,
                    264,
                    163,
                    227,
                    238,
                    490,
                    83,
                    276,
                    69,
                    323,
                    169,
                    252,
                    253,
                    884,
                    70,
                    772,
                    480,
                    295,
                    911,
                    17,
                    2,
                ],
            ],
            dtype=torch.int32,
        ),
    )
    torch_assertclose(
        pieces.attention_mask().bool_mask.squeeze(dim=(1, 2)),
        torch.tensor(
            [
                [
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ],
                [
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ],
                [
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                ],
            ]
        ),
    )



import warnings
from importlib.util import find_spec

_has_scipy = find_spec("scipy") is not None and find_spec("scipy.stats") is not None
has_bitsandbytes = find_spec("bitsandbytes") is not None

# As of v0.40.0, `bitsandbytes` doesn't correctly specify `scipy` as an installation
# dependency. This can lead to situations where the former is installed but
# the latter isn't and the ImportError gets masked. So, we additionally check
# for the presence of `scipy`.
if has_bitsandbytes and not _has_scipy:
    warnings.warn(
        "The `bitsandbytes` library is installed but its dependency "
        "`scipy` isn't. Please install `scipy` to correctly load `bitsandbytes`."
    )
    has_bitsandbytes = False

has_safetensors = find_spec("safetensors")



from typing import Optional

import torch
from torch import Tensor

from ...layers.transformer import (
    EmbeddingDropouts,
    EmbeddingLayerNorms,
    TransformerEmbeddings,
)


class RoBERTaEmbeddings(TransformerEmbeddings):
    """
    RoBERTa (`Liu et al., 2019`_) embedding layer.

    This layer only differs from
    :py:class:`~curated_transformers.layers.transformer.TransformerEmbeddingsLayer`
    in the computation of positions.

    .. _Liu et al., 2019: https://arxiv.org/abs/1907.11692
    """

    def __init__(
        self,
        *,
        dropouts: EmbeddingDropouts,
        embedding_width: int,
        hidden_width: int,
        layer_norms: EmbeddingLayerNorms,
        n_pieces: int,
        n_positions: Optional[int],
        n_types: Optional[int],
        padding_id: int,
        device: Optional[torch.device] = None,
    ):
        """
        Construct a RoBERTa embeddings layer.

        :param dropouts:
            Dropouts to use in the embeddings layer.
        :param embedding_width:
            Width of the embeddings.
        :param hidden_width:
            Hidden width of the transformer. If this width differs from
            ``embedding_width``, a projection layer is added to ensure
            that the output of the embeddings layer has the same width as the
            transformer.
        :param layer_norms:
            Layer norms to use in the embeddings layer.
        :param n_pieces:
            Number of piece embeddings.
        :param n_positions:
            Number of position embeddings. Position embeddings are disabled
            by using ``None``. Position embeddings can be used to inform the
            model of input order.
        :param n_types:
            Number of type embeddings. Type embeddings are disabled by using
            ``None``. Type embeddings can be used to inform the model of the
            spans of different sequences in the input.
        :param padding_id:
            Padding identifier. The padding identifier is used to compute the
            positions when they are not provided to the ``forward`` method.
        :param device:
            Device on which the module is to be initialized.
        """

        super().__init__(
            dropouts=dropouts,
            embedding_width=embedding_width,
            hidden_width=hidden_width,
            layer_norms=layer_norms,
            n_pieces=n_pieces,
            n_positions=n_positions,
            n_types=n_types,
            device=device,
        )

        self.padding_id = padding_id

    def _get_positions(self, x: Tensor) -> Tensor:
        # We need to generate the position IDs from the
        # input tensor to pass to the embedding layer and
        # handle padding, c.f https://github.com/huggingface/transformers/blob/330247ede2d8265aae9ab0b7a0d1a811c344960d/src/transformers/models/roberta/modeling_roberta.py#L1566

        mask = x.ne(self.padding_id).int()
        return (mask.cumsum(dim=1) * mask) + self.padding_id

    def forward(
        self,
        piece_ids: Tensor,
        *,
        positions: Optional[Tensor] = None,
        type_ids: Optional[Tensor] = None,
    ) -> Tensor:
        if positions is None:
            positions = self._get_positions(piece_ids)
        return super().forward(
            piece_ids,
            positions=positions,
            type_ids=type_ids,
        )



from abc import ABC, abstractmethod
from typing import Any, Dict, Mapping, Optional, Type, TypeVar

from fsspec import AbstractFileSystem
from huggingface_hub.utils import EntryNotFoundError

from ..repository.file import RepositoryFile
from ..repository.fsspec import FsspecArgs, FsspecRepository
from ..repository.hf_hub import HfHubRepository
from ..repository.repository import Repository, TokenizerRepository

SelfFromHF = TypeVar("SelfFromHF", bound="FromHF")


class FromHF(ABC):
    """
    Mixin class for downloading tokenizers from Hugging Face Hub.

    It directly queries the Hugging Face Hub to load the tokenizer from
    its configuration file.
    """

    @classmethod
    @abstractmethod
    def from_hf_hub_to_cache(
        cls: Type[SelfFromHF],
        *,
        name: str,
        revision: str = "main",
    ):
        """
        Download the tokenizer's serialized model, configuration and vocab files
        from Hugging Face Hub into the local Hugging Face cache directory.
        Subsequent loading of the tokenizer will read the files from disk. If the
        files are already cached, this is a no-op.

        :param name:
            Model name.
        :param revision:
            Model revision.
        """
        raise NotImplementedError

    @classmethod
    def from_fsspec(
        cls: Type[SelfFromHF],
        *,
        fs: AbstractFileSystem,
        model_path: str,
        fsspec_args: Optional[FsspecArgs] = None,
    ) -> SelfFromHF:
        """
        Construct a tokenizer and load its parameters from an fsspec filesystem.

        :param fs:
            Filesystem.
        :param model_path:
            The model path.
        :param fsspec_args:
            Implementation-specific keyword arguments to pass to fsspec
            filesystem operations.
        :returns:
            The tokenizer.
        """
        return cls.from_repo(
            repo=FsspecRepository(fs, model_path, fsspec_args),
        )

    @classmethod
    def from_hf_hub(
        cls: Type[SelfFromHF], *, name: str, revision: str = "main"
    ) -> SelfFromHF:
        """
        Construct a tokenizer and load its parameters from Hugging Face Hub.

        :param name:
            Model name.
        :param revision:
            Model revision.
        :returns:
            The tokenizer.
        """
        return cls.from_repo(
            repo=HfHubRepository(name=name, revision=revision),
        )

    @classmethod
    @abstractmethod
    def from_repo(
        cls: Type[SelfFromHF],
        repo: Repository,
    ) -> SelfFromHF:
        """
        Construct and load a tokenizer from a repository.

        :param repository:
            The repository to load from.
        :returns:
            Loaded tokenizer.
        """
        ...


SelfLegacyFromHF = TypeVar("SelfLegacyFromHF", bound="LegacyFromHF")


class LegacyFromHF(FromHF):
    """
    Subclass of :class:`.FromHF` for legacy tokenizers. This subclass
    implements the ``from_hf_hub`` method and provides through the abstract
    ``_load_from_vocab_files`` method:

    * The vocabulary files requested by a tokenizer through the
    ``vocab_files`` member variable.
    * The tokenizer configuration (when available).
    """

    vocab_files: Dict[str, str] = {}

    @classmethod
    @abstractmethod
    def _load_from_vocab_files(
        cls: Type[SelfLegacyFromHF],
        *,
        vocab_files: Mapping[str, RepositoryFile],
        tokenizer_config: Optional[Dict[str, Any]],
    ) -> SelfLegacyFromHF:
        """
        Construct a tokenizer from its vocabulary files and optional
        configuration.

        :param vocab_files:
            The resolved vocabulary files (in a local cache).
        :param tokenizer_config:
            The tokenizer configuration (when available).
        :returns:
            The tokenizer.
        """
        raise NotImplementedError

    @classmethod
    def from_hf_hub_to_cache(
        cls: Type[SelfLegacyFromHF],
        *,
        name: str,
        revision: str = "main",
    ):
        repo = TokenizerRepository(HfHubRepository(name, revision=revision))
        for _, filename in cls.vocab_files.items():
            _ = repo.file(filename)

        try:
            _ = repo.tokenizer_config()
        except EntryNotFoundError:
            pass

    @classmethod
    def from_repo(
        cls: Type[SelfLegacyFromHF],
        repo: Repository,
    ) -> SelfLegacyFromHF:
        repo = TokenizerRepository(repo)
        vocab_files = {}
        for vocab_file, filename in cls.vocab_files.items():
            vocab_files[vocab_file] = repo.file(filename)

        try:
            tokenizer_config = repo.tokenizer_config()
        except OSError:
            tokenizer_config = None

        return cls._load_from_vocab_files(
            vocab_files=vocab_files, tokenizer_config=tokenizer_config
        )



import pytest
import torch

from curated_transformers.layers.attention import AttentionMask
from curated_transformers.models.falcon._hf import HFConfigKeys
from curated_transformers.models.falcon.decoder import FalconDecoder
from curated_transformers.models.hf_hub.conversion import CommonHFKeys

from ...compat import has_hf_transformers, has_torch_compile
from ...conftest import TORCH_DEVICES
from ...utils import torch_assertclose
from ..util import (
    JITMethod,
    assert_decoder_output_equals_hf,
    assert_model_hf_serialization_roundtrip,
)

N_PIECES = 1024

# We do not have tests to check caching/positions against upstream, there
# are two issues with the upstream model:
#
# 1. It uses the torch scaled dot-product attention, but that has a bug
#    in generating causal masks:
#
#    https://github.com/pytorch/pytorch/issues/103082
#
# 2. When using a cache, the upstream implementation does not take into
#    account that we need to index by position into the rotary embeddings.
#
# We test caching instead by comparing output of the model with caching
# against output without caching.


FALCON_TEST_MODELS = [
    (
        "explosion-testing/falcon-no-parallel-attn-test",
        "f9122de6cafad10159af214786fa11b89cc37a89",
    ),
    ("explosion-testing/falcon-test", "24ff3d5fd83b4d174888356f20e61349f6cbf467"),
    (
        "explosion-testing/refined-web-model-test",
        "57a7a9829a4b6fce833152c4c20a46c7056f9cc1",
    ),
    (
        "explosion-testing/refined-web-model-new-decoder-test",
        "512fb26ab864280eace45144924b2c213ad87a87",
    ),
    # Enable when HF transformers with Falcon is released.
    # (
    #    "explosion-testing/falcon-new-decoder-test",
    #    "d53e01af2cc0edc22719c4d2f22bd66a87fa8c64",
    # ),
    # Enable when HF transformers with Falcon is released.
    # (
    #     "explosion-testing/falcon-new-decoder-alibi-test",
    #     "a6a0d422ba272a4395eb4da0b831cfdc0c571f82",
    # ),
]


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
@pytest.mark.parametrize("model_revision", FALCON_TEST_MODELS)
@pytest.mark.parametrize("with_torch_sdp", [False, True])
def test_decoder(torch_device, model_revision, with_torch_sdp):
    model, revision = model_revision
    assert_decoder_output_equals_hf(
        FalconDecoder,
        model,
        torch_device,
        model_revision=revision,
        trust_remote_code=True,
        with_cache=False,
        with_mask=False,
        with_positions=False,
        with_torch_sdp=with_torch_sdp,
    )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.skipif(not has_torch_compile, reason="requires torch.compile")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
@pytest.mark.parametrize("model_revision", FALCON_TEST_MODELS)
@pytest.mark.parametrize("with_torch_sdp", [False, True])
def test_decoder_with_torch_compile(torch_device, model_revision, with_torch_sdp):
    model, revision = model_revision
    assert_decoder_output_equals_hf(
        FalconDecoder,
        model,
        torch_device,
        model_revision=revision,
        trust_remote_code=True,
        with_cache=False,
        with_mask=False,
        with_positions=False,
        jit_method=JITMethod.TorchCompile,
        with_torch_sdp=with_torch_sdp,
    )


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
@pytest.mark.parametrize("model_revision", FALCON_TEST_MODELS)
def test_decoder_with_cache(torch_device, model_revision):
    model, revision = model_revision

    model = FalconDecoder.from_hf_hub(
        name=model, revision=revision, device=torch_device
    )
    model.eval()

    torch.manual_seed(0)
    X = torch.randint(0, N_PIECES, (2, 10), device=torch_device)
    mask = torch.ones_like(X, dtype=torch.bool)
    X_rest = torch.randint(0, N_PIECES, (2, 10), device=torch_device)
    mask_rest = torch.cat([mask, torch.ones_like(X_rest, dtype=torch.bool)], dim=1)

    with torch.no_grad():
        Y = model(X, AttentionMask(mask), store_cache=True)
        Y = model(
            X_rest, AttentionMask(mask_rest), cache=Y.cache
        ).last_hidden_layer_state
        Y_no_cache = model(
            torch.cat([X, X_rest], dim=1), AttentionMask(mask_rest)
        ).last_hidden_layer_state

    torch_assertclose(Y, Y_no_cache[:, 10:, :])


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
def test_decoder_hf_serializtion_roundtrip(torch_device):
    # We only support round-trip serialization for the non-RWM models.
    assert_model_hf_serialization_roundtrip(
        FalconDecoder,
        "explosion-testing/falcon-test",
        torch_device,
        model_revision="24ff3d5fd83b4d174888356f20e61349f6cbf467",
        trust_remote_code=True,
        optional_hf_config_keys={
            HFConfigKeys.N_HEAD_KV.name,
            HFConfigKeys.NUM_HEAD_KV.name,
            HFConfigKeys.MULTI_QUERY.name,
            HFConfigKeys.N_HEAD_KV.name,
            HFConfigKeys.NEW_DECODER_ARCHITECTURE.name,
            CommonHFKeys.ATTENTION_PROBS_DROPOUT_PROB.name,
            CommonHFKeys.HIDDEN_DROPOUT_PROB.name,
        },
    )



from typing import Any, Dict, Iterable, List, Mapping, Optional, Tuple, Type, TypeVar

from curated_tokenizers import ByteBPEProcessor

from ...repository.file import RepositoryFile
from ..hf_hub import LegacyFromHF
from ..util import remove_pieces_from_sequence
from .bbpe_tokenizer import ByteBPETokenizer
from .legacy_tokenizer import AddBosEosPreEncoder, PreDecoder

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="RoBERTaTokenizer")


class RoBERTaPreDecoder(PreDecoder):
    def __init__(
        self,
        *,
        bos_id: int,
        eos_id: int,
    ):
        """
        Construct a RoBERTa pre-decoder.

        :param bos_id:
            The piece id used to mark the beginning of a sequence.
        :param eos_id:
            The piece id used to mark the end of a sequence.
        """
        self.bos_id = bos_id
        self.eos_id = eos_id

    def __call__(self, input: Iterable[Iterable[int]]) -> List[List[int]]:
        return [
            list(remove_pieces_from_sequence(ids, (self.bos_id, self.eos_id)))
            for ids in input
        ]


class RoBERTaTokenizer(ByteBPETokenizer, LegacyFromHF):
    """
    Legacy tokenizer for RoBERTa (`Liu et al., 2019`_) models.

    .. _Liu et al., 2019: https://arxiv.org/abs/1907.11692
    """

    vocab_files: Dict[str, str] = {"vocab": "vocab.json", "merges": "merges.txt"}

    def __init__(
        self,
        *,
        vocab: Dict[str, int],
        merges: List[Tuple[str, str]],
        special_pieces: Optional[Dict[str, int]] = None,
        bos_piece: str = "<s>",
        eos_piece: str = "</s>",
    ):
        """
        Construct a RoBERTa tokenizer.

        :param vocab:
            The word piece vocabulary.
        :param merges:
            Merges.
        :param special_pieces:
            Special pieces.
        :param bos_piece:
            Beginning of sequence piece.
        :param eos_piece:
            End of sequence piece.
        """
        super().__init__(vocab=vocab, merges=merges, special_pieces=special_pieces)

        bos_id = _get_piece_id_or_fail(self.processor, bos_piece)
        eos_id = _get_piece_id_or_fail(self.processor, eos_piece)

        self._eos_piece = eos_piece

        self.pre_decoder = RoBERTaPreDecoder(
            bos_id=bos_id,
            eos_id=eos_id,
        )

        self.pre_encoder = AddBosEosPreEncoder(bos_piece=bos_piece, eos_piece=eos_piece)

    @classmethod
    def from_files(
        cls: Type[Self],
        *,
        vocab_file: RepositoryFile,
        merges_file: RepositoryFile,
        bos_piece: str = "<s>",
        eos_piece: str = "</s>",
    ) -> Self:
        """
        Construct a tokenizer from vocabulary and merge files.

        :param vocab_file:
            The vocabulary file.
        :param merges_file:
            The merges file.
        :param bos_piece:
            The piece to use to mark the beginning of a sequence.
        :param eos_piece:
            The piece to use to mark the end of a sequence.
        """
        with vocab_file.open(mode="r", encoding="utf-8") as vocab:
            with merges_file.open(mode="r", encoding="utf-8") as merges:
                processor = ByteBPEProcessor.load_from_files(vocab=vocab, merges=merges)
        return cls(
            # This is a bit annoying, but we want to avoid these extremely
            # overloaded constructors.
            vocab=processor.vocab,
            merges=processor.merges,
            bos_piece=bos_piece,
            eos_piece=eos_piece,
        )

    @property
    def eos_piece(self) -> Optional[str]:
        return self._eos_piece

    @classmethod
    def _load_from_vocab_files(
        cls: Type[Self],
        *,
        vocab_files: Mapping[str, RepositoryFile],
        tokenizer_config: Optional[Dict[str, Any]],
    ) -> Self:
        return cls.from_files(
            vocab_file=vocab_files["vocab"], merges_file=vocab_files["merges"]
        )


def _get_piece_id_or_fail(processor: ByteBPEProcessor, piece: str):
    piece_id = processor.piece_to_id(piece)
    if piece_id is None:
        raise ValueError(
            f"RoBERTa piece encoder vocabulary doesn't contain '{piece}' piece"
        )
    return piece_id



from typing import Any, Dict, List, Mapping, Optional, Tuple

import torch

from ...util.string import StringTransform, StringTransformations
from ..hf_hub.conversion import (
    CommonHFKeys,
    HFConfigKey,
    HFConfigKeyDefault,
    HFSpecificConfig,
    config_from_hf,
    config_to_hf,
)
from ..module import DecoderModule
from .config import GPTNeoXConfig

# Order-dependent.
COMMON_HF_PARAM_KEY_TRANSFORMS: List[StringTransform] = [
    StringTransformations.sub("gpt_neox", "decoder"),
    # Attention blocks.
    StringTransformations.sub(".attention", ".mha"),
    StringTransformations.sub(".mha.query_key_value", ".mha.input"),
    StringTransformations.sub(".mha.dense", ".mha.output"),
    # Pointwise feedforward.
    StringTransformations.sub(".mlp", ".ffn"),
    StringTransformations.sub(".dense_h_to_4h", ".intermediate"),
    StringTransformations.sub(".ffn.dense_4h_to_h", ".ffn.output"),
    # Layer norms.
    StringTransformations.sub(".input_layernorm", ".attn_input_layer_norm"),
    StringTransformations.sub(".post_attention_layernorm", ".ffn_input_layer_norm"),
    StringTransformations.sub("final_layer_norm.", "output_layer_norm."),
    # Embeddings.
    StringTransformations.sub("embed_in.", "embeddings.piece_embeddings."),
    StringTransformations.sub("embed_out.", "output_embeddings."),
]

DECODER_HF_PARAM_KEY_TRANSFORMS = [
    StringTransformations.remove_prefix("gpt_neox.", reversible=False)
] + COMMON_HF_PARAM_KEY_TRANSFORMS
CAUSAL_LM_HF_PARAM_KEY_TRANSFORMS = COMMON_HF_PARAM_KEY_TRANSFORMS


class HFConfigKeys:
    @staticmethod
    def conv_rotary_embedding_base(config: GPTNeoXConfig) -> int:
        assert config.layer.attention.rotary_embeddings is not None
        return config.layer.attention.rotary_embeddings.rotary_base

    @staticmethod
    def conv_rotary_embedding_fraction(config: GPTNeoXConfig) -> float:
        assert config.layer.attention.rotary_embeddings is not None
        return config.layer.attention.rotary_embeddings.rotary_fraction

    ROTARY_EMB_BASE = HFConfigKey(
        "rotary_emb_base",
        "rotary_embedding_base",
        lambda c: HFConfigKeys.conv_rotary_embedding_base(c),
    )
    ROTARY_PCT = HFConfigKey(
        "rotary_pct",
        "rotary_embedding_fraction",
        lambda c: HFConfigKeys.conv_rotary_embedding_fraction(c),
    )


HF_CONFIG_KEYS: List[Tuple[HFConfigKey, Optional[HFConfigKeyDefault]]] = [
    (CommonHFKeys.DTYPE, HFConfigKeyDefault("float16")),
    (CommonHFKeys.HIDDEN_ACT, None),
    (CommonHFKeys.HIDDEN_SIZE, None),
    (CommonHFKeys.INTERMEDIATE_SIZE, None),
    (CommonHFKeys.LAYER_NORM_EPS, None),
    (CommonHFKeys.VOCAB_SIZE, None),
    (CommonHFKeys.MAX_POSITION_EMBEDDINGS, None),
    (CommonHFKeys.NUM_HIDDEN_LAYERS, None),
    (CommonHFKeys.NUM_ATTENTION_HEADS_UNIFORM, None),
    (HFConfigKeys.ROTARY_EMB_BASE, None),
    (HFConfigKeys.ROTARY_PCT, None),
    (CommonHFKeys.ATTENTION_PROBS_DROPOUT_PROB, HFConfigKeyDefault(0.0)),
    (CommonHFKeys.HIDDEN_DROPOUT_PROB, HFConfigKeyDefault(0.0)),
]

HF_SPECIFIC_CONFIG_DECODER = HFSpecificConfig(
    architectures=["GPTNeoXModel"], model_type="gpt_neox"
)
HF_SPECIFIC_CONFIG_CAUSAL_LM = HFSpecificConfig(
    architectures=["GPTNeoXForCausalLM"], model_type="gpt_neox"
)


def _config_from_hf(hf_config: Mapping[str, Any]) -> GPTNeoXConfig:
    kwargs = config_from_hf("GPT-NeoX", hf_config, HF_CONFIG_KEYS)

    return GPTNeoXConfig(
        model_max_length=CommonHFKeys.MAX_POSITION_EMBEDDINGS.get_kwarg(kwargs),
        **kwargs,
    )


def _config_to_hf(cls, curated_config: GPTNeoXConfig) -> Dict[str, Any]:
    out = config_to_hf(curated_config, [k for k, _ in HF_CONFIG_KEYS])
    if issubclass(cls, DecoderModule):
        return HF_SPECIFIC_CONFIG_DECODER.merge(out)
    else:
        return HF_SPECIFIC_CONFIG_CAUSAL_LM.merge(out)



from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, Mapping, Optional, Type, TypeVar

import torch
from fsspec import AbstractFileSystem
from torch import Tensor
from torch.nn import Module

from ...quantization import prepare_module_for_quantization
from ...quantization.bnb.config import BitsAndBytesConfig
from ...repository.fsspec import FsspecArgs, FsspecRepository
from ...repository.hf_hub import HfHubRepository
from ...repository.repository import ModelRepository, Repository
from ...util.serde.load import load_model_from_checkpoints
from ..module import ConfigT, TransformerModule

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="FromHF")


class FromHF(ABC, Generic[ConfigT]):
    """
    Mixin class for downloading models from Hugging Face Hub.

    Implementation of the mixin's abstract methods will provide various ``from_``
    methods to load a model, including the ``from_hf_hub`` method to download a
    model from the Hugging Face Hub.
    """

    @classmethod
    @abstractmethod
    def state_dict_from_hf(cls, params: Mapping[str, Tensor]) -> Mapping[str, Tensor]:
        """
        Convert a state dict of a Hugging Face model to a valid
        state dict for the module.

        :param params:
            The state dict to convert.
        :returns:
            The converted state dict.
        """
        ...

    @classmethod
    @abstractmethod
    def state_dict_to_hf(cls, params: Mapping[str, Tensor]) -> Mapping[str, Tensor]:
        """
        Convert the state dict of the module to a compatible
        Hugging Face model's format.

        :param params:
            The state dict to convert.
        :returns:
            The converted state dict.
        """
        ...

    @classmethod
    @abstractmethod
    def config_from_hf(cls, hf_config: Mapping[str, Any]) -> ConfigT:
        """
        Convert a Hugging Face model configuration to the
        module's configuration.

        :param hf_config:
            The Hugging Face model configuration.
        :returns:
            The converted Curated Transformer
            configuration.
        """
        ...

    @classmethod
    @abstractmethod
    def config_to_hf(cls, curated_config: ConfigT) -> Mapping[str, Any]:
        """
        Convert the module's  configuration to the a
        Hugging Face model configuration.

        :param curated_config:
            The Curated Transformer model configuration.
        :returns:
            The converted Hugging Face configuration.
        """
        ...

    @classmethod
    @abstractmethod
    def from_hf_config(
        cls: Type[Self],
        *,
        hf_config: Any,
        device: Optional[torch.device] = None,
    ) -> Self:
        """
        Create the module from a Hugging Face model JSON-deserialized
        model configuration.

        :param hf_config:
            Hugging Face model configuration.
        :param device:
            Device on which to initialize the model.
        :returns:
            Module constructed using the configuration.
        """
        ...

    @classmethod
    def from_hf_hub_to_cache(
        cls: Type[Self],
        *,
        name: str,
        revision: str = "main",
    ):
        """
        Download the model's weights from Hugging Face Hub into the local
        Hugging Face cache directory. Subsequent loading of the
        model will read the weights from disk. If the weights are already
        cached, this is a no-op.

        :param name:
            Model name.
        :param revision:
            Model revision.
        """
        repo = ModelRepository(HfHubRepository(name=name, revision=revision))
        repo.model_config()
        repo.model_checkpoints()

    @classmethod
    def from_fsspec(
        cls: Type[Self],
        *,
        fs: AbstractFileSystem,
        model_path: str,
        fsspec_args: Optional[FsspecArgs] = None,
        device: Optional[torch.device] = None,
        quantization_config: Optional[BitsAndBytesConfig] = None,
    ) -> Self:
        """
        Construct a module and load its parameters from a fsspec filesystem.

        :param fs:
            The filesystem to load the model from.
        :param model_path:
            The path of the model on the filesystem.
        :param fsspec_args:
            Implementation-specific keyword arguments to pass to fsspec
            filesystem operations.
        :param device:
            Device on which the model is initialized.
        :param quantization_config:
            Configuration for loading quantized weights.
        :returns:
            Module with the parameters loaded.
        """
        return cls.from_repo(
            repo=FsspecRepository(fs, model_path, fsspec_args),
            device=device,
            quantization_config=quantization_config,
        )

    def from_fsspec_(
        self: Self,
        *,
        fs: AbstractFileSystem,
        model_path: str,
        fsspec_args: Optional[FsspecArgs] = None,
        device: Optional[torch.device] = None,
        quantization_config: Optional[BitsAndBytesConfig] = None,
    ) -> Self:
        """
        Load parameters from a fsspec filestytem in-place into the model.

        :param fs:
            The filesystem to load the model from.
        :param model_path:
            The path of the model on the filesystem.
        :param fsspec_args:
            Implementation-specific keyword arguments to pass to fsspec
            filesystem operations.
        :param device:
            Device on which the model is initialized.
        :param quantization_config:
            Configuration for loading quantized weights.
        :returns:
            Module with the parameters loaded.
        """
        return self.from_repo_(
            repo=FsspecRepository(fs, model_path, fsspec_args),
            device=device,
            quantization_config=quantization_config,
        )

    @classmethod
    def from_hf_hub(
        cls: Type[Self],
        *,
        name: str,
        revision: str = "main",
        device: Optional[torch.device] = None,
        quantization_config: Optional[BitsAndBytesConfig] = None,
    ) -> Self:
        """
        Construct a module and load its parameters from Hugging Face Hub.

        :param name:
            Model name.
        :param revision:
            Model revision.
        :param device:
            Device on which the model is initialized.
        :param quantization_config:
            Configuration for loading quantized weights.
        :returns:
            Module with the parameters loaded.
        """
        return cls.from_repo(
            repo=HfHubRepository(name=name, revision=revision),
            device=device,
            quantization_config=quantization_config,
        )

    def from_hf_hub_(
        self: Self,
        *,
        name: str,
        revision: str = "main",
        device: Optional[torch.device] = None,
        quantization_config: Optional[BitsAndBytesConfig] = None,
    ) -> Self:
        """
        Load parameters from Hugging Face Hub in-place into the model.

        :param name:
            Model name.
        :param revision:
            Model revision.
        :param device:
            Device on which the model is initialized.
        :param quantization_config:
            Configuration for loading quantized weights.
        :returns:
            Module with the parameters loaded.
        """
        return self.from_repo_(
            repo=HfHubRepository(name=name, revision=revision),
            device=device,
            quantization_config=quantization_config,
        )

    @classmethod
    @abstractmethod
    def is_supported(cls: Type[Self], config: Dict[str, Any]) -> bool:
        """
        Check if the model with the given configuration is supported by this
        class.

        :param config:
            Hugging Face model configuration.
        :returns:
            Whether the model is supported by this class.
        """
        raise NotImplementedError

    @classmethod
    def from_repo(
        cls: Type[Self],
        *,
        repo: Repository,
        device: Optional[torch.device] = None,
        quantization_config: Optional[BitsAndBytesConfig] = None,
    ) -> Self:
        """
        Construct and load a model from a repository.

        :param repository:
            The repository to load from.
        :param device:
            Device on which to initialize the model.
        :param quantization_config:
            Configuration for loading quantized weights.
        :returns:
            Loaded model.
        """
        model_repo = ModelRepository(repo)
        config = model_repo.model_config()
        model = cls.from_hf_config(hf_config=config, device=torch.device("meta"))
        assert isinstance(model, Module)

        return model.from_repo_(
            repo=repo, device=device, quantization_config=quantization_config
        )

    def from_repo_(
        self: Self,
        *,
        repo: Repository,
        device: Optional[torch.device] = None,
        quantization_config: Optional[BitsAndBytesConfig] = None,
    ) -> Self:
        """
        Load parameters from a repository in-place into the model.

        :param repository:
            The repository to load from.
        :param device:
            Device on which to initialize the model.
        :param quantization_config:
            Configuration for loading quantized weights.
        :returns:
            Loaded model.
        """
        model_repo = ModelRepository(repo)

        # Convert the model to the expected dtype.
        assert isinstance(self, TransformerModule)
        dtype: torch.dtype = self.config.dtype
        self.to(dtype=dtype)

        # Prepare for quantization.
        if quantization_config is not None:
            tensor2param = prepare_module_for_quantization(self, quantization_config)  # type: ignore
        else:
            tensor2param = None

        # Download model and convert HF parameter names to ours.
        checkpoint_filenames, checkpoint_type = model_repo.model_checkpoints()
        load_model_from_checkpoints(
            self,  # type:ignore
            filepaths=checkpoint_filenames,
            checkpoint_type=checkpoint_type,
            state_dict_converter=type(self).state_dict_from_hf,
            tensor_to_param_converter=tensor2param,
            device=device,
        )

        # Ensure that any non-persistent buffers are also moved to
        # the correct device.
        if device is not None:
            self.to(device)

        return self



from dataclasses import dataclass
from typing import Callable, List, Optional

from torch.nn import Module


@dataclass
class ModuleIterator:
    """
    Represents the details of a module when travesing a PyTorch module graph.

    :param module:
        Current module.
    :param name:
        Name of the module.
    :param prefix:
        Current dot path of the module. Includes the name.
    :param parent:
        Parent module. Will be `None` for the root module.
    """

    module: Module
    name: str
    prefix: str
    parent: Optional[Module]


def apply_to_module(module: Module, func: Callable[[ModuleIterator], None]):
    """
    Apply a function the module and its submodules in a breadth-first
    fashion.

    :param module:
        Root module.
    :param func:
        A callable that takes a module iterator as its argument.
    """
    queue: List[ModuleIterator] = [ModuleIterator(module, "", "", None)]

    while queue:
        itr = queue.pop(0)
        func(itr)

        for name, child in itr.module.named_children():
            child_prefix = f"{itr.prefix}.{name}" if itr.prefix else name
            queue.append(ModuleIterator(child, name, child_prefix, itr.module))
